__all__ = ("spectres",)

import warnings

import numpy as np

# Taken from https://github.com/ACCarnall/SpectRes
# (which is under GPL 3), cite https://arxiv.org/abs/1705.05165


def make_bins(wavs):
    """Given a series of wavelength points, find the edges and widths
    of corresponding wavelength bins."""
    edges = np.zeros(wavs.shape[0] + 1)
    widths = np.zeros(wavs.shape[0])
    edges[0] = wavs[0] - (wavs[1] - wavs[0]) / 2
    widths[-1] = wavs[-1] - wavs[-2]
    edges[-1] = wavs[-1] + (wavs[-1] - wavs[-2]) / 2
    edges[1:-1] = (wavs[1:] + wavs[:-1]) / 2
    widths[:-1] = edges[1:-1] - edges[:-2]

    return edges, widths


def spectres(new_wavs, spec_wavs, spec_fluxes, spec_errs=None, fill=0, verbose=True):
    """
    Function for resampling spectra (and optionally associated
    uncertainties) onto a new wavelength basis.

    If performance is an issue, https://github.com/ACCarnall/SpectRes
    includes a version that is wrapped in numba and should be much faster.

    Parameters
    ----------

    new_wavs : numpy.ndarray
        Array containing the new wavelength sampling desired for the
        spectrum or spectra.

    spec_wavs : numpy.ndarray
        1D array containing the current wavelength sampling of the
        spectrum or spectra.

    spec_fluxes : numpy.ndarray
        Array containing spectral fluxes at the wavelengths specified in
        spec_wavs, last dimension must correspond to the shape of
        spec_wavs. Extra dimensions before this may be used to include
        multiple spectra.

    spec_errs : numpy.ndarray (optional)
        Array of the same shape as spec_fluxes containing uncertainties
        associated with each spectral flux value.

    fill : float (optional)
        Where new_wavs extends outside the wavelength range in spec_wavs
        this value will be used as a filler in new_fluxes and new_errs.

    verbose : bool (optional)
        Setting verbose to False will suppress the default warning about
        new_wavs extending outside spec_wavs and "fill" being used.

    Returns
    -------

    new_fluxes : numpy.ndarray
        Array of resampled flux values, last dimension is the same
        length as new_wavs, other dimensions are the same as
        spec_fluxes.

    new_errs : numpy.ndarray
        Array of uncertainties associated with fluxes in new_fluxes.
        Only returned if spec_errs was specified.
    """

    # Rename the input variables for clarity within the function.
    old_wavs = spec_wavs
    old_fluxes = spec_fluxes
    old_errs = spec_errs

    # Make arrays of edge positions and widths for the old and new bins

    old_edges, old_widths = make_bins(old_wavs)
    new_edges, new_widths = make_bins(new_wavs)

    # Generate output arrays to be populated
    new_fluxes = np.zeros(old_fluxes[..., 0].shape + new_wavs.shape)

    if old_errs is not None:
        if old_errs.shape != old_fluxes.shape:
            raise ValueError("If specified, spec_errs must be the same shape " "as spec_fluxes.")
        else:
            new_errs = np.copy(new_fluxes)

    start = 0
    stop = 0

    # Calculate new flux and uncertainty values, looping over new bins
    for j in range(new_wavs.shape[0]):

        # Add filler values if new_wavs extends outside of spec_wavs
        if (new_edges[j] < old_edges[0]) or (new_edges[j + 1] > old_edges[-1]):
            new_fluxes[..., j] = fill

            if spec_errs is not None:
                new_errs[..., j] = fill

            if (j == 0 or j == new_wavs.shape[0] - 1) and verbose:
                warnings.warn(
                    "Spectres: new_wavs contains values outside the range "
                    "in spec_wavs, new_fluxes and new_errs will be filled "
                    "with the value set in the 'fill' keyword argument "
                    "(by default 0).",
                    category=RuntimeWarning,
                )
            continue

        # Find first old bin which is partially covered by the new bin
        while old_edges[start + 1] <= new_edges[j]:
            start += 1

        # Find last old bin which is partially covered by the new bin
        while old_edges[stop + 1] < new_edges[j + 1]:
            stop += 1

        # If new bin is fully inside an old bin start and stop are equal
        if stop == start:
            new_fluxes[..., j] = old_fluxes[..., start]
            if old_errs is not None:
                new_errs[..., j] = old_errs[..., start]

        # Otherwise multiply the first and last old bin widths by P_ij
        else:
            start_factor = (old_edges[start + 1] - new_edges[j]) / (old_edges[start + 1] - old_edges[start])

            end_factor = (new_edges[j + 1] - old_edges[stop]) / (old_edges[stop + 1] - old_edges[stop])

            old_widths[start] *= start_factor
            old_widths[stop] *= end_factor

            # Populate new_fluxes spectrum and uncertainty arrays
            f_widths = old_widths[start : stop + 1] * old_fluxes[..., start : stop + 1]
            new_fluxes[..., j] = np.sum(f_widths, axis=-1)
            new_fluxes[..., j] /= np.sum(old_widths[start : stop + 1])

            if old_errs is not None:
                e_wid = old_widths[start : stop + 1] * old_errs[..., start : stop + 1]

                new_errs[..., j] = np.sqrt(np.sum(e_wid**2, axis=-1))
                new_errs[..., j] /= np.sum(old_widths[start : stop + 1])

            # Put back the old bin widths to their initial values
            old_widths[start] /= start_factor
            old_widths[stop] /= end_factor

    # If errors were supplied return both new_fluxes and new_errs.
    if old_errs is not None:
        return new_fluxes, new_errs

    # Otherwise just return the new_fluxes spectrum array
    else:
        return new_fluxes



import unittest
import warnings

import numpy as np

from rubin_sim.maf.slicers import OneDSlicer, UniSlicer


def make_data_values(size=100, min_val=0.0, max_val=1.0, random=-1):
    """Generate a simple array of numbers, evenly arranged between min/max,
    but (optional) random order."""
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(max_val) - float(min_val)) / (datavalues.max() - datavalues.min())
    datavalues += min_val
    if random > 0:
        rng = np.random.RandomState(random)
        randorder = rng.rand(size)
        randind = np.argsort(randorder)
        datavalues = datavalues[randind]
    datavalues = np.array(list(zip(datavalues)), dtype=[("testdata", "float")])
    return datavalues


class TestOneDSlicerSetup(unittest.TestCase):
    def setUp(self):
        self.testslicer = OneDSlicer(slice_col_name="testdata")

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_slicertype(self):
        """Test instantiation of slicer sets slicer type as expected."""
        self.assertEqual(self.testslicer.slicer_name, self.testslicer.__class__.__name__)
        self.assertEqual(self.testslicer.slicer_name, "OneDSlicer")

    def test_setup_slicer_bins(self):
        """Test setting up slicer using defined bins."""
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        bins = np.arange(dvmin, dvmax + 0.05, 0.1)
        dv = make_data_values(nvalues, dvmin, dvmax, random=4)
        # Used right bins?
        self.testslicer = OneDSlicer(slice_col_name="testdata", bins=bins)
        self.testslicer.setup_slicer(dv)
        np.testing.assert_equal(self.testslicer.bins, bins)
        self.assertEqual(self.testslicer.nslice, len(bins) - 1)

    def test_setup_slicer_nbins_zeros(self):
        """Test what happens if give slicer test data that is all
        single-value."""
        dv = np.zeros(100, float)
        dv = np.array(list(zip(dv)), dtype=[("testdata", "float")])
        bin_size = 0.1
        self.testslicer = OneDSlicer(slice_col_name="testdata", bin_size=bin_size)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.testslicer.setup_slicer(dv)
            self.assertIn("creasing bin_max", str(w[-1].message))
        self.assertEqual(self.testslicer.nslice, 2)

    def test_setup_slicer_limits(self):
        """Test setting up slicer using bin_min/max."""
        bin_min = 0
        bin_max = 1
        dvmin = -0.5
        dvmax = 1.5
        dv = make_data_values(1000, dvmin, dvmax, random=342)
        self.testslicer = OneDSlicer(slice_col_name="testdata", bin_min=bin_min, bin_max=bin_max)
        self.testslicer.setup_slicer(dv)
        self.assertAlmostEqual(self.testslicer.bins.min(), bin_min)
        self.assertAlmostEqual(self.testslicer.bins.max(), bin_max)

    def test_setup_slicer_binsize(self):
        """Test setting up slicer using bin_size."""
        dvmin = 0
        dvmax = 1
        dv = make_data_values(1000, dvmin, dvmax, random=8977)
        # Test basic use.
        bin_size = 0.5
        self.testslicer = OneDSlicer(slice_col_name="testdata", bin_size=bin_size)
        self.testslicer.setup_slicer(dv)
        self.assertEqual(self.testslicer.nslice, (dvmax - dvmin) / bin_size)

    def test_setup_slicer_freedman(self):
        """Test that setting up the slicer using bins=None works."""
        dvmin = 0
        dvmax = 1
        dv = make_data_values(1000, dvmin, dvmax, random=2234)
        self.testslicer = OneDSlicer(slice_col_name="testdata", bins=None)
        self.testslicer.setup_slicer(dv)
        # How many bins do you expect from optimal bin_size?
        from rubin_sim.maf.utils import optimal_bins

        bins = optimal_bins(dv["testdata"])
        np.testing.assert_equal(self.testslicer.nslice, bins)


class TestOneDSlicerIteration(unittest.TestCase):
    def setUp(self):
        self.testslicer = OneDSlicer(slice_col_name="testdata")
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        self.bins = np.arange(dvmin, dvmax, 0.01)
        dv = make_data_values(nvalues, dvmin, dvmax, random=5678)
        self.testslicer = OneDSlicer(slice_col_name="testdata", bins=self.bins)
        self.testslicer.setup_slicer(dv)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_iteration(self):
        """Test iteration."""
        for i, s in enumerate(self.testslicer):
            self.assertEqual(s["slice_point"]["sid"], i)
            self.assertEqual(s["slice_point"]["bin_left"], self.bins[i])
            self.assertEqual(s["slice_point"]["bin_right"], self.bins[i + 1])
        # Iteration stops while i<= nslice ..
        # nslice = len(bins)-1 (and bins[i+2] is out of range
        # but bins[i+1] is last RH edge of bins
        self.assertEqual(i, self.testslicer.nslice - 1)

    def test_get_item(self):
        """Test that can return an individual indexed values of the slicer."""
        for i in [0, 10, 20]:
            self.assertEqual(self.testslicer[i]["slice_point"]["sid"], i)
            self.assertEqual(self.testslicer[i]["slice_point"]["bin_left"], self.bins[i])


class TestOneDSlicerEqual(unittest.TestCase):
    def setUp(self):
        self.testslicer = OneDSlicer(slice_col_name="testdata")

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_equivalence(self):
        """Test equals method."""
        # Note that two OneD slicers will be considered equal if
        # they are both the same kind of slicer AND have the same bins.
        # Set up self..
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        bins = np.arange(dvmin, dvmax, 0.01)
        dv = make_data_values(nvalues, dvmin, dvmax, random=32499)
        # Set up reference
        self.testslicer = OneDSlicer(slice_col_name="testdata", bins=bins)
        self.testslicer.setup_slicer(dv)
        # Set up another slicer to match
        # (same bins, although not the same data).
        dv2 = make_data_values(nvalues + 100, dvmin, dvmax, random=334)
        testslicer2 = OneDSlicer(slice_col_name="testdata", bins=bins)
        testslicer2.setup_slicer(dv2)
        self.assertTrue(self.testslicer == testslicer2)
        self.assertFalse(self.testslicer != testslicer2)
        # Set up another slicer that should not match
        # (different bins, based on data)
        dv2 = make_data_values(nvalues, dvmin + 1, dvmax + 1, random=445)
        testslicer2 = OneDSlicer(slice_col_name="testdata", bins=None)
        testslicer2.setup_slicer(dv2)
        self.assertTrue(self.testslicer != testslicer2)
        self.assertFalse(self.testslicer == testslicer2)
        # Set up a different kind of slicer that should not match.
        dv2 = make_data_values(100, 0, 1, random=12)
        testslicer2 = UniSlicer()
        testslicer2.setup_slicer(dv2)
        self.assertTrue(self.testslicer != testslicer2)
        self.assertFalse(self.testslicer == testslicer2)
        # Get another oneDslicer that is not set up, and check equivalence.
        testslicer2 = OneDSlicer(slice_col_name="testdata")
        self.assertTrue(self.testslicer != testslicer2)
        self.assertFalse(self.testslicer == testslicer2)
        testslicer2 = OneDSlicer(slice_col_name="testdata", bin_min=0, bin_max=1, bin_size=0.5)
        testslicer3 = OneDSlicer(slice_col_name="testdata", bin_min=0, bin_max=1, bin_size=0.5)
        self.assertTrue(testslicer2 == testslicer3)
        self.assertFalse(testslicer2 != testslicer3)
        testslicer3 = OneDSlicer(slice_col_name="testdata", bin_min=0, bin_max=1)
        self.assertFalse(testslicer2 == testslicer3)
        self.assertTrue(testslicer2 != testslicer3)
        usebins = np.arange(0, 1, 0.1)
        testslicer2 = OneDSlicer(slice_col_name="testdata", bins=usebins)
        testslicer3 = OneDSlicer(slice_col_name="testdata", bins=usebins)
        self.assertTrue(testslicer2 == testslicer3)
        self.assertFalse(testslicer2 != testslicer3)
        testslicer3 = OneDSlicer(slice_col_name="testdata", bins=usebins + 1)
        self.assertFalse(testslicer2 == testslicer3)
        self.assertTrue(testslicer2 != testslicer3)


class TestOneDSlicerSlicing(unittest.TestCase):
    long_message = True

    def setUp(self):
        self.testslicer = OneDSlicer(slice_col_name="testdata")

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_slicing(self):
        """Test slicing."""
        dvmin = 0
        dvmax = 1
        stepsize = 0.01
        bins = np.arange(dvmin, dvmax + stepsize / 2, stepsize)
        # Test that testbinner raises appropriate error before it's set up
        # (first time)
        self.assertRaises(NotImplementedError, self.testslicer._slice_sim_data, 0)
        # test that we're not dumping extra visits into the 'last bin'
        # (that it's closed both sides)
        nvalues = 1000
        dv = make_data_values(nvalues, 0, 2, random=560)
        bins = np.arange(0, 1.05, 0.1)
        self.testslicer = OneDSlicer(slice_col_name="testdata", bins=bins)
        nbins = len(bins) - 1
        self.testslicer.setup_slicer(dv)
        for i, s in enumerate(self.testslicer):
            self.assertEqual(len(s["idxs"]), nvalues / float(nbins) / 2.0)

    def test_night_slicing(self):
        """Test slicing with 'night' column"""
        numval = 1000
        vals = np.arange(0, numval, 1) / (numval / 50)
        nights = np.floor(vals)
        np.random.shuffle(nights)
        # If you were doing by-hand comparison, this can be useful
        # bins = np.arange(0, 50.5, 1)
        # vals, bb = np.histogram(nights, bins=bins)
        vals = make_data_values(len(nights), random=10)
        vals = np.array(list(zip(nights, vals["testdata"])), dtype=[("night", "int"), ("testdata", "float")])
        test_slicer = OneDSlicer(slice_col_name="night", bins=np.arange(0, 50.5, 1))
        test_slicer.setup_slicer(vals)
        for s in test_slicer:
            self.assertTrue(vals[s["idxs"]]["night"].min() >= s["slice_point"]["bin_left"])
            self.assertTrue(vals[s["idxs"]]["night"].max() < s["slice_point"]["bin_right"])
            self.assertEqual(len(s["idxs"]), numval / 50)


if __name__ == "__main__":
    unittest.main()



"""
Created on Tue Sep 18 13:35:41 2018

@author: rstreet
"""

__all__ = ("calculate_lsst_field_visibility", "calculate_lsst_field_visibility_fast")


import astropy.units as u
import numpy as np
from astropy.coordinates import AltAz, EarthLocation, SkyCoord, get_sun
from astropy.time import Time, TimeDelta
from rubin_scheduler.utils import Site, approx_ra_dec2_alt_az


def calculate_lsst_field_visibility(
    ra, dec, start_date, end_date, min_alt=30.0, sun_alt_limit=18.0, sample_rate=0.0007, verbose=False
):
    """Method to calculate the visibility of a given RA and Dec from LSST
    over the course of a year

    Adapted from an example in the Astropy docs.

    Parameters
    ----------
    ra : `float`
        RA in decimal degrees.
    dec : `float`
        Declination in decimal degrees
    start_date : `astropy.time.Time`
        Start date for calculations
    end_date : `astropy.time.Time`
        End date for calculations
    min_alt : `float`, optional
        Minimal altitude for field
    sun_alt_limit : `float`, optional
        Maximum sun altitude to consider for visibility
    sample_rate : `float`, optional
        Time spacing between visibility tests (days)
    verbose : `bool`, optional
        Output extra information, including debugging
    """
    field = SkyCoord(ra, dec, frame="icrs", unit=(u.deg, u.deg))

    lsst_site = Site("LSST")
    lsst = EarthLocation(
        lat=lsst_site.latitude * u.deg,
        lon=lsst_site.longitude * u.deg,
        height=lsst_site.height * u.m,
    )

    total_time_visible = 0.0

    dates = np.arange(start_date, end_date, TimeDelta(1, format="jd", scale="tai"))

    target_alts = []
    hrs_visible_per_night = []
    hrs_per_night = []

    for d in dates:
        intervals = np.arange(0.0, 1.0, sample_rate)

        dt = TimeDelta(intervals, format="jd", scale="tai")

        ts = d + dt

        frame = AltAz(obstime=ts, location=lsst)

        altaz = field.transform_to(frame)

        alts = np.array((altaz.alt * u.deg).value)

        idx = np.where(alts >= min_alt)[0]

        sun_altaz = get_sun(ts).transform_to(frame)

        sun_alts = np.array((sun_altaz.alt * u.deg).value)

        jdx = np.where(sun_alts < sun_alt_limit)[0]
        # Hours available in the sun
        hrs_per_night.append(sample_rate * len(sun_alts[jdx]) * 24.0)
        # The indexes where sun down and target above min_alt
        idx = list(set(idx).intersection(set(jdx)))
        # The highest altitude for the target, when the sun is down
        target_alts.append(alts[jdx].max())

        if len(idx) > 0:
            ts_vis = ts[idx]

            tvis = sample_rate * len(ts_vis)

            total_time_visible += tvis

            if verbose:
                print("Target visible from LSST for " + str(round(tvis * 24.0, 2)) + "hrs on " + d.isot)
            # Hours of visibility of the target on this night
            hrs_visible_per_night.append((tvis * 24.0))

        else:
            # target_alts.append(-1e5)
            hrs_visible_per_night.append(0.0)

            if verbose:
                print("Target not visible from LSST on " + d.isot)

    return np.array(total_time_visible), np.array(hrs_visible_per_night)


def calculate_lsst_field_visibility_fast(
    ra,
    dec,
    start_date,
    end_date,
    min_alt=30.0,
    sun_alt_limit=18.0,
    sample_rate=0.0007,
):
    """Method to calculate the visibility of a given RA and Dec from LSST
    over the course of a year

    Skips astropy calculation of alt/az which is slow and uses
    approximate transform from rubin_scheduler.

    Parameters
    ----------
    ra : `float`
        RA in decimal degrees.
    dec : `float`
        Declination in decimal degrees
    start_date : `astropy.time.Time`
        Start date for calculations
    end_date : `astropy.time.Time`
        End date for calculations
    min_alt : `float`, optional
        Minimal altitude for field
    sun_alt_limit : `float`, optional
        Maximum sun altitude to consider for visibility
    cadence : `float`, optional
        Time spacing between visibility tests (days)

    Returns
    -------
    tvisible : `float`
        Total time target is visible (days)
    dates_visible : `np.ndarray`, (N,)
        Dates that target is above min_alt and sun is below sun_alt_limit,
        within start_date to end_date.
    """
    lsst_site = Site("LSST")

    dates = np.arange(start_date.mjd, end_date.mjd + sample_rate / 2.0, sample_rate)

    alts, _ = approx_ra_dec2_alt_az(ra, dec, lsst_site.latitude, lsst_site.longitude, dates)
    # where is the target above the minimum altitude
    target_high = np.where(alts >= min_alt)[0]

    # when is the sun above the sun_alt_limit
    dates = dates[target_high]
    sun_locs = get_sun(Time(dates, format="mjd", scale="utc"))
    sun_alts, _ = approx_ra_dec2_alt_az(
        sun_locs.ra.deg, sun_locs.dec.deg, lsst_site.latitude, lsst_site.longitude, dates
    )
    sun_low = np.where(sun_alts <= sun_alt_limit)[0]

    dates = dates[sun_low]
    # total amount of time target is visible, in days
    tvisible = len(dates) * sample_rate

    return tvisible, dates



__all__ = ("NgalScaleMetric", "NlcPointsMetric")

import healpy as hp
import numpy as np

from rubin_sim.maf.maps import TrilegalDensityMap
from rubin_sim.phot_utils import DustValues

from .base_metric import BaseMetric


class NgalScaleMetric(BaseMetric):
    """Approximate number of galaxies, scaled by median seeing.

    Parameters
    ----------
    a_max : `float`
        The maximum dust extinction to allow. Anything with higher dust
        extinction is considered to have zero usable galaxies.
    m5min : `float`
        The minimum coadded 5-sigma depth to allow. Anything less is
        considered to have zero usable galaxies.
    filter : `str`
        The filter to use. Any visits in other filters are ignored.
    """

    def __init__(
        self,
        seeing_col="seeingFwhmEff",
        m5_col="fiveSigmaDepth",
        metric_name="NgalScale",
        filtername="i",
        a_max=0.2,
        m5min=26.0,
        filter_col="filter",
        **kwargs,
    ):
        maps = ["DustMap"]
        units = "N gals"
        self.seeing_col = seeing_col
        self.m5_col = m5_col
        self.filtername = filtername
        self.filter_col = filter_col
        self.a_max = a_max
        self.m5min = m5min

        super().__init__(
            col=[self.m5_col, self.filter_col, self.seeing_col],
            maps=maps,
            units=units,
            metric_name=metric_name,
            **kwargs,
        )
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

    def run(self, data_slice, slice_point):
        # I'm a little confused why there's a dust cut and an M5 cut,
        # but whatever
        a_x = self.ax1[data_slice[self.filter_col][0]] * slice_point["ebv"]
        if a_x > self.a_max:
            return 0

        in_filt = np.where(data_slice[self.filter_col] == self.filtername)
        coadd_m5 = 1.25 * np.log10(np.sum(10.0 ** (0.8 * data_slice[self.m5_col][in_filt])))
        if coadd_m5 < self.m5min:
            return 0

        theta = np.median(data_slice[self.seeing_col])
        # N gals per arcmin2
        ngal_per_arcmin2 = 57 * (0.75 / theta) ** 1.5

        area = hp.nside2pixarea(slice_point["nside"], degrees=True) * 3600.0

        ngal = ngal_per_arcmin2 * area
        return ngal


class NlcPointsMetric(BaseMetric):
    """Number of points in stellar light curves

    Parameters
    ----------
    ndpmin : `int`
        The number of points to demand on a lightcurve in a single
        filter to have that light curve qualify.
    mags : `float`
        The magnitude of our fiducial object (maybe make it a dict in the
        future to support arbitrary colors).
    maps : `list` [`~rubin_sim.maf.map`] or None
        List of stellar density maps to use.
        Default of None loads Trilegal maps.
    nside : `int`
        The nside is needed to make sure the loaded maps
        match the slicer nside.
    """

    def __init__(
        self,
        ndpmin=10,
        mags=21.0,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        metric_name="NlcPoints",
        maps=None,
        nside=128,
        **kwargs,
    ):
        units = "N LC points"
        self.ndpmin = ndpmin
        self.mags = mags
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.nside = nside
        if maps is None:
            maps = [TrilegalDensityMap(filtername=fn, nside=nside) for fn in "ugrizy"]
        super().__init__(
            col=[self.m5_col, self.filter_col],
            maps=maps,
            units=units,
            metric_name=metric_name,
            **kwargs,
        )

    def run(self, data_slice, slice_point):
        if self.nside != slice_point["nside"]:
            raise ValueError("nside of metric does not match slicer")
        pix_area = hp.nside2pixarea(slice_point["nside"], degrees=True)

        nlcpoints = 0
        # Let's do it per filter
        for filtername in np.unique(data_slice[self.filter_col]):
            in_filt = np.where(
                (data_slice[self.filter_col] == filtername) & (data_slice[self.m5_col] > self.mags)
            )[0]
            n_obs = np.size(in_filt)
            if n_obs > self.ndpmin:
                nstars = (
                    np.interp(
                        self.mags,
                        slice_point[f"starMapBins_{filtername}"][1:],
                        slice_point[f"starLumFunc_{filtername}"],
                    )
                    * pix_area
                )
                nlcpoints += n_obs * nstars

        return nlcpoints



__all__ = (
    "id2intid",
    "intid2id",
    "load_spec_files",
    "BaseSingleInterp",
    "ScatteredStar",
    "LowerAtm",
    "UpperAtm",
    "MergedSpec",
    "Airglow",
    "TwilightInterp",
    "MoonInterp",
    "ZodiacalInterp",
)

import glob
import os
import warnings

import healpy as hp
import numpy as np
from rubin_scheduler.data import get_data_dir
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d

from rubin_sim.phot_utils import Bandpass, Sed

from .twilight_func import twilight_func

# Make backwards compatible with healpy
if hasattr(hp, "get_interp_weights"):
    get_neighbours = hp.get_interp_weights
elif hasattr(hp, "get_neighbours"):
    get_neighbours = hp.get_neighbours
else:
    print("Could not find appropriate healpy function for get_interp_weight or get_neighbours")


def id2intid(ids):
    """Take an array of ids, and convert them to an integer id.
    Handy if you want to put things into a sparse array.
    """
    uids = np.unique(ids)
    order = np.argsort(ids)
    oids = ids[order]
    uintids = np.arange(np.size(uids), dtype=int)
    left = np.searchsorted(oids, uids)
    right = np.searchsorted(oids, uids, side="right")
    intids = np.empty(ids.size, dtype=int)
    for i in range(np.size(left)):
        intids[left[i] : right[i]] = uintids[i]
    result = intids * 0
    result[order] = intids
    return result, uids, uintids


def intid2id(intids, uintids, uids, dtype=int):
    """convert an int back to an id"""
    ids = np.zeros(np.size(intids))

    order = np.argsort(intids)
    ointids = intids[order]
    left = np.searchsorted(ointids, uintids, side="left")
    right = np.searchsorted(ointids, uintids, side="right")
    for i, (le, ri) in enumerate(zip(left, right)):
        ids[le:ri] = uids[i]
    result = np.zeros(np.size(intids), dtype=dtype)
    result[order] = ids

    return result


def load_spec_files(filenames, mags=False):
    """Load up the ESO spectra.

    The ESO npz files contain the following arrays:
    * filter_wave: The central wavelengths of the pre-computed magnitudes
    * wave: wavelengths for the spectra
    * spec: array of spectra and magnitudes along with the relevant
    variable inputs.  For example, airglow has dtype =
    [('airmass', '<f8'),
    ('solar_flux', '<f8'),
    ('spectra','<f8', (17001,)),
    ('mags', '<f8', (6,)]

    For each unique airmass and solar_flux value, there is a 17001 elements
    spectra and 6 magnitudes.
    """

    if len(filenames) == 1:
        temp = np.load(filenames[0])
        wave = temp["wave"].copy()
        # Note old camelCase here. Might need to update if sav
        # files regenerated
        filter_wave = temp["filterWave"].copy()
        if mags:
            # don't copy the spectra to save memory space
            dt = np.dtype(
                [
                    (key, temp["spec"].dtype[i])
                    for i, key in enumerate(temp["spec"].dtype.names)
                    if key != "spectra"
                ]
            )
            spec = np.zeros(temp["spec"].size, dtype=dt)
            for key in temp["spec"].dtype.names:
                if key != "spectra":
                    spec[key] = temp["spec"][key].copy()
        else:
            spec = temp["spec"].copy()
    else:
        temp = np.load(filenames[0])
        wave = temp["wave"].copy()
        filter_wave = temp["filterWave"].copy()
        if mags:
            # don't copy the spectra to save memory space
            dt = np.dtype(
                [
                    (key, temp["spec"].dtype[i])
                    for i, key in enumerate(temp["spec"].dtype.names)
                    if key != "spectra"
                ]
            )
            spec = np.zeros(temp["spec"].size, dtype=dt)
            for key in temp["spec"].dtype.names:
                if key != "spectra":
                    spec[key] = temp["spec"][key].copy()
        else:
            spec = temp["spec"].copy()
        for filename in filenames[1:]:
            temp = np.load(filename)
            if mags:
                # don't copy the spectra to save memory space
                dt = np.dtype(
                    [
                        (key, temp["spec"].dtype[i])
                        for i, key in enumerate(temp["spec"].dtype.names)
                        if key != "spectra"
                    ]
                )
                tempspec = np.zeros(temp["spec"].size, dtype=dt)
                for key in temp["spec"].dtype.names:
                    if key != "spectra":
                        tempspec[key] = temp["spec"][key].copy()
            else:
                tempspec = temp["spec"]
            spec = np.append(spec, tempspec)
    return spec, wave, filter_wave


class BaseSingleInterp:
    """Base class for interpolating sky components which only depend
    on airmass.

    Parameters
    ----------
    comp_name : `str`, optional
        Component name.
    sorted_order : `list` [`str`], optional
        Order of the dimensions in the input .npz files.
    mags : `bool`, optional
        Return magnitudes (only) rather than the full spectrum.
    """

    def __init__(self, comp_name=None, sorted_order=["airmass", "nightTimes"], mags=False):
        self.mags = mags

        data_dir = os.path.join(get_data_dir(), "skybrightness", "ESO_Spectra/" + comp_name)

        filenames = sorted(glob.glob(data_dir + "/*.npz"))
        self.spec, self.wave, self.filter_wave = load_spec_files(filenames, mags=self.mags)

        # Take the log of the spectra in case we want to interp in log space.
        if not mags:
            self.log_spec = np.zeros(self.spec["spectra"].shape, dtype=float)
            good = np.where(self.spec["spectra"] != 0)
            self.log_spec[good] = np.log10(self.spec["spectra"][good])
            self.spec_size = self.spec["spectra"][0].size
        else:
            self.spec_size = 0

        # What order are the dimensions sorted by (from how
        # the .npz was packaged)
        self.sorted_order = sorted_order
        self.dim_dict = {}
        self.dim_sizes = {}
        for dt in self.sorted_order:
            self.dim_dict[dt] = np.unique(self.spec[dt])
            self.dim_sizes[dt] = np.size(np.unique(self.spec[dt]))

        # Set up and save the dict to order the filters once.
        self.filter_name_dict = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}

    def __call__(self, interp_points, filter_names=["u", "g", "r", "i", "z", "y"]):
        """At `interp_points (e.g. airmass), return values."""
        if self.mags:
            return self.interp_mag(interp_points, filter_names=filter_names)
        else:
            return self.interp_spec(interp_points)

    def _indx_and_weights(self, points, grid):
        """For given 1-D points, find the grid points on
        either side and return the weights assume grid is sorted.

        Parameters
        ----------
        points : `np.ndarray`, (N,)
            The points on the grid to query.
        grid : `np.ndarray`, (N,)
            The grid on which to locate `points`.

        Returns
        -------
        indx_r, indx_l : `np.ndarray`, `np.ndarray`
            The grid indexes for each of the 1-d points
        w_r, w_l : `np.ndarray`, `np.ndarray`
            The weights for each of these grid points.
        """

        order = np.argsort(points)

        indx_r = np.empty(points.size, dtype=int)

        indx_r[order] = np.searchsorted(grid, points[order])
        indx_l = indx_r - 1

        # If points off the grid were requested, just use the edge grid point
        off_grid = np.where(indx_r == grid.size)
        indx_r[off_grid] = grid.size - 1
        full_range = grid[indx_r] - grid[indx_l]

        w_l = np.zeros(full_range.size, dtype=float)
        w_r = np.ones(full_range.size, dtype=float)

        good = np.where(full_range != 0)
        w_l[good] = (grid[indx_r][good] - points[good]) / full_range[good]
        w_r[good] = (points[good] - grid[indx_l[good]]) / full_range[good]

        return indx_r, indx_l, w_r, w_l

    def _weighting(self, interp_points, values):
        """
        given a list/array of airmass values, return a
        dict with the interpolated spectrum at each airmass
        and the wavelength array.

        Input interp_points should be sorted
        """
        results = np.zeros((interp_points.size, np.size(values[0])), dtype=float)

        in_range = np.where(
            (interp_points["airmass"] <= np.max(self.dim_dict["airmass"]))
            & (interp_points["airmass"] >= np.min(self.dim_dict["airmass"]))
        )
        indx_r, indx_l, w_r, w_l = self._indx_and_weights(
            interp_points["airmass"][in_range], self.dim_dict["airmass"]
        )

        nextra = 1

        # XXX--should I use the log spectra?
        # Make a check and switch back and forth?
        results[in_range] = (
            w_r[:, np.newaxis] * values[indx_r * nextra] + w_l[:, np.newaxis] * values[indx_l * nextra]
        )

        return results

    def interp_spec(self, interp_points):
        result = self._weighting(interp_points, self.log_spec)
        mask = np.where(result == 0.0)
        result = 10.0**result
        result[mask] = 0.0
        return {"spec": result, "wave": self.wave}

    def interp_mag(self, interp_points, filter_names=["u", "g", "r", "i", "z", "y"]):
        filterindx = [self.filter_name_dict[key] for key in filter_names]
        result = self._weighting(interp_points, self.spec["mags"][:, filterindx])
        mask = np.where(result == 0.0)
        result = 10.0 ** (-0.4 * (result - np.log10(3631.0)))
        result[mask] = 0.0
        return {"spec": result, "wave": self.filter_wave}


class ScatteredStar(BaseSingleInterp):
    """Interpolate the spectra caused by scattered starlight."""

    def __init__(self, comp_name="ScatteredStarLight", mags=False):
        super(ScatteredStar, self).__init__(comp_name=comp_name, mags=mags)


class LowerAtm(BaseSingleInterp):
    """Interpolate the spectra caused by the lower atmosphere."""

    def __init__(self, comp_name="LowerAtm", mags=False):
        super(LowerAtm, self).__init__(comp_name=comp_name, mags=mags)


class UpperAtm(BaseSingleInterp):
    """Interpolate the spectra caused by the upper atmosphere."""

    def __init__(self, comp_name="UpperAtm", mags=False):
        super(UpperAtm, self).__init__(comp_name=comp_name, mags=mags)


class MergedSpec(BaseSingleInterp):
    """Interpolate the combined spectra caused by the sum of the scattered
    starlight, air glow, upper and lower atmosphere.
    """

    def __init__(self, comp_name="MergedSpec", mags=False):
        super(MergedSpec, self).__init__(comp_name=comp_name, mags=mags)


class Airglow(BaseSingleInterp):
    """Interpolate the spectra caused by airglow."""

    def __init__(self, comp_name="Airglow", sorted_order=["airmass", "solarFlux"], mags=False):
        super(Airglow, self).__init__(comp_name=comp_name, mags=mags, sorted_order=sorted_order)
        self.n_solar_flux = np.size(self.dim_dict["solarFlux"])

    def _weighting(self, interp_points, values):
        results = np.zeros((interp_points.size, np.size(values[0])), dtype=float)
        # Only interpolate point that lie in the model grid
        in_range = np.where(
            (interp_points["airmass"] <= np.max(self.dim_dict["airmass"]))
            & (interp_points["airmass"] >= np.min(self.dim_dict["airmass"]))
            & (interp_points["solar_flux"] >= np.min(self.dim_dict["solarFlux"]))
            & (interp_points["solar_flux"] <= np.max(self.dim_dict["solarFlux"]))
        )
        use_points = interp_points[in_range]
        am_right_index, am_left_index, am_right_w, am_left_w = self._indx_and_weights(
            use_points["airmass"], self.dim_dict["airmass"]
        )

        sf_right_index, sf_left_index, sf_right_w, sf_left_w = self._indx_and_weights(
            use_points["solar_flux"], self.dim_dict["solarFlux"]
        )

        for am_index, amW in zip([am_right_index, am_left_index], [am_right_w, am_left_w]):
            for sf_index, sfW in zip([sf_right_index, sf_left_index], [sf_right_w, sf_left_w]):
                results[in_range] += (
                    amW[:, np.newaxis] * sfW[:, np.newaxis] * values[am_index * self.n_solar_flux + sf_index]
                )
        return results


class TwilightInterp:
    """Use the Solar Spectrum to provide an interpolated spectra or magnitudes
    for the twilight sky.

    Parameters
    ----------
    mags : `bool`
        If True, only return the LSST filter magnitudes,
        otherwise return the full spectrum
    dark_sky_mags : `dict`
        Dict of the zenith dark sky values to be assumed.
        The twilight fits are done relative to the dark sky level.
    fit_results : `dict`
        Dict of twilight parameters based on twilight_func.
        Keys should be filter names.
    """

    def __init__(self, mags=False, dark_sky_mags=None, fit_results=None):
        if dark_sky_mags is None:
            dark_sky_mags = {
                "u": 22.8,
                "g": 22.3,
                "r": 21.2,
                "i": 20.3,
                "z": 19.3,
                "y": 18.0,
                "B": 22.35,
                "G": 21.71,
                "R": 21.3,
            }

        self.mags = mags

        data_dir = os.path.join(get_data_dir(), "skybrightness")

        solar_saved = np.load(os.path.join(data_dir, "solarSpec/solarSpec.npz"))
        self.solar_spec = Sed(wavelen=solar_saved["wave"], flambda=solar_saved["spec"])
        solar_saved.close()

        canon_filters = {}
        fnames = ["blue_canon.csv", "green_canon.csv", "red_canon.csv"]

        # Filter names, from bluest to reddest.
        self.filter_names = ["B", "G", "R"]

        # Supress warning that Canon filters are low sampling
        warnings.filterwarnings("ignore", message="There is an area of")
        warnings.filterwarnings("ignore", message="Wavelength sampling of")
        for fname, filter_name in zip(fnames, self.filter_names):
            bpdata = np.genfromtxt(
                os.path.join(data_dir, "Canon/", fname),
                delimiter=", ",
                dtype=list(zip(["wave", "through"], [float] * 2)),
            )
            bp_temp = Bandpass()
            bp_temp.set_bandpass(bpdata["wave"], bpdata["through"])
            bp_temp.resample_bandpass(
                wavelen_min=self.solar_spec.wavelen.min(),
                wavelen_max=self.solar_spec.wavelen.max(),
                wavelen_step=self.solar_spec.wavelen[1] - self.solar_spec.wavelen[0],
            )
            # Force wavelengths to be identical so
            # it doesn't try to resample again later
            bp_temp.wavelen = self.solar_spec.wavelen
            canon_filters[filter_name] = bp_temp

        # Tack on the LSST filters
        through_path = os.path.join(get_data_dir(), "throughputs", "baseline")
        lsst_keys = ["u", "g", "r", "i", "z", "y"]
        for key in lsst_keys:
            bp = np.loadtxt(
                os.path.join(through_path, "total_" + key + ".dat"),
                dtype=list(zip(["wave", "trans"], [float] * 2)),
            )
            temp_b = Bandpass()
            temp_b.set_bandpass(bp["wave"], bp["trans"])
            canon_filters[key] = temp_b
            self.filter_names.append(key)

        # MAGIC NUMBERS from fitting the all-sky camera:
        # Code to generate values in
        # sims_skybrightness/examples/fitTwiSlopesSimul.py
        # Which in turn uses twilight maps from
        # sims_skybrightness/examples/buildTwilMaps.py
        # values are of the form:
        # 0: ratio of f^z_12 to f_dark^z
        # 1: slope of curve wrt sun alt
        # 2: airmass term (10^(arg[2]*(X-1)))
        # 3: azimuth term.
        # 4: zenith dark sky flux (erg/s/cm^2)

        # For z and y, just assuming the shape parameter
        # fits are similar to the other bands.
        # Looks like the diode is not sensitive enough to detect faint sky.
        # Using the Patat et al 2006 I-band values for z and
        # modeified a little for y as a temp fix.
        if fit_results is None:
            self.fit_results = {
                "B": [
                    7.56765633e00,
                    2.29798055e01,
                    2.86879956e-01,
                    3.01162143e-01,
                    2.58462036e-04,
                ],
                "G": [
                    2.38561156e00,
                    2.29310648e01,
                    2.97733083e-01,
                    3.16403197e-01,
                    7.29660095e-04,
                ],
                "R": [
                    1.75498017e00,
                    2.22011802e01,
                    2.98619033e-01,
                    3.28880254e-01,
                    3.24411056e-04,
                ],
                "z": [2.29, 24.08, 0.3, 0.3, -666],
                "y": [2.0, 24.08, 0.3, 0.3, -666],
            }

            # XXX-completely arbitrary fudge factor to make things
            # brighter in the blue
            # Just copy the blue and say it's brighter.
            self.fit_results["u"] = [
                16.0,
                2.29622121e01,
                2.85862729e-01,
                2.99902574e-01,
                2.32325117e-04,
            ]
        else:
            self.fit_results = fit_results

        # Take out any filters that don't have fit results
        self.filter_names = [key for key in self.filter_names if key in self.fit_results]

        self.eff_wave = []
        self.solar_mag = []
        for filter_name in self.filter_names:
            self.eff_wave.append(canon_filters[filter_name].calc_eff_wavelen()[0])
            self.solar_mag.append(self.solar_spec.calc_mag(canon_filters[filter_name]))

        order = np.argsort(self.eff_wave)
        self.filter_names = np.array(self.filter_names)[order]
        self.eff_wave = np.array(self.eff_wave)[order]
        self.solar_mag = np.array(self.solar_mag)[order]

        # update the fit results to be zeropointed properly
        for key in self.fit_results:
            f0 = 10.0 ** (-0.4 * (dark_sky_mags[key] - np.log10(3631.0)))
            self.fit_results[key][-1] = f0

        self.solar_wave = self.solar_spec.wavelen
        self.solar_flux = self.solar_spec.flambda
        # This one isn't as bad as the model grids, maybe we could get
        # away with computing the magnitudes in the __call__ each time.
        if mags:
            # Load up the LSST filters and convert the
            # solarSpec.flambda and solarSpec.wavelen to fluxes
            self.lsst_filter_names = ["u", "g", "r", "i", "z", "y"]
            self.lsst_equations = np.zeros(
                (np.size(self.lsst_filter_names), np.size(self.fit_results["B"])),
                dtype=float,
            )
            self.lsst_eff_wave = []

            fits = np.empty((np.size(self.eff_wave), np.size(self.fit_results["B"])), dtype=float)
            for i, fn in enumerate(self.filter_names):
                fits[i, :] = self.fit_results[fn]

            through_path = os.path.join(get_data_dir(), "throughputs", "baseline")
            for filtername in self.lsst_filter_names:
                bp = np.loadtxt(
                    os.path.join(through_path, "total_" + filtername + ".dat"),
                    dtype=list(zip(["wave", "trans"], [float] * 2)),
                )
                temp_b = Bandpass()
                temp_b.set_bandpass(bp["wave"], bp["trans"])
                self.lsst_eff_wave.append(temp_b.calc_eff_wavelen()[0])
            # Loop through the parameters and interpolate to new
            # eff wavelengths
            for i in np.arange(self.lsst_equations[0, :].size):
                interp = InterpolatedUnivariateSpline(self.eff_wave, fits[:, i])
                self.lsst_equations[:, i] = interp(self.lsst_eff_wave)
            # Set the dark sky flux
            for i, filter_name in enumerate(self.lsst_filter_names):
                self.lsst_equations[i, -1] = 10.0 ** (-0.4 * (dark_sky_mags[filter_name] - np.log10(3631.0)))

        self.filter_name_dict = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}

    def print_fits_used(self):
        """Print out the fit parameters being used"""
        print(
            r"\\tablehead{\colhead{Filter} & \colhead{$r_{12/z}$} & "
            r"\colhead{$a$ (1/radians)} & \colhead{$b$ (1/airmass)} & "
            r"\colhead{$c$ (az term/airmass)} & "
            r"\colhead{$f_z_dark$ (erg/s/cm$^2$)$\\times 10^8$} & "
            r"\colhead{m$_z_dark$}}"
        )
        for key in self.fit_results:
            numbers = ""
            for num in self.fit_results[key]:
                if num > 0.001:
                    numbers += " & %.2f" % num
                else:
                    numbers += " & %.2f" % (num * 1e8)
            print(
                key,
                numbers,
                " & ",
                "%.2f" % (-2.5 * np.log10(self.fit_results[key][-1]) + np.log10(3631.0)),
            )

    def __call__(self, intep_points, filter_names=["u", "g", "r", "i", "z", "y"]):
        if self.mags:
            return self.interp_mag(intep_points, filter_names=filter_names)
        else:
            return self.interp_spec(intep_points)

    def interp_mag(
        self,
        interp_points,
        max_am=3.0,
        limits=(np.radians(15.0), np.radians(-20.0)),
        filter_names=["u", "g", "r", "i", "z", "y"],
    ):
        """
        Parameters
        ----------
        interp_points : `np.ndarray`, (N, 3)
            Interpolation points. Should contain sunAlt, airmass and azRelSun.
        max_am : `float`, optional
            Maximum airmass to calculate twilight sky to.
        limits : `np.ndarray`, (N,), optional
            Sun altitude limits

        Returns
        -------
        spectra, wavelength : `np.ndarray`, (N, 3), `np.ndarray`, (M,)

        Note
        ----
        Originally fit the twilight with a cutoff of sun altitude of
        -11 degrees. I think it can be safely extrapolated farther,
        but be warned you may be entering a regime where it breaks down.
        """
        npts = len(filter_names)
        result = np.zeros((np.size(interp_points), npts), dtype=float)

        out_of_range = np.where(interp_points["sunAlt"] > np.radians(-11))[0]
        if np.size(out_of_range) > 0:
            warnings.warn("Extrapolating twilight beyond a sun altitude of -11 degrees")

        good = np.where(
            (interp_points["sunAlt"] >= np.min(limits))
            & (interp_points["sunAlt"] <= np.max(limits))
            & (interp_points["airmass"] <= max_am)
            & (interp_points["airmass"] >= 1.0)
        )[0]

        for i, filterName in enumerate(filter_names):
            out_of_range = np.where(interp_points["sunAlt"] > np.max(limits))[0]
            if np.size(out_of_range) > 0:
                result[:, i] = np.nan
            else:
                result[good, i] = twilight_func(
                    interp_points[good], *self.lsst_equations[self.filter_name_dict[filterName], :].tolist()
                )

        return {"spec": result, "wave": self.lsst_eff_wave}

    def interp_spec(self, interp_points, max_am=3.0, limits=(np.radians(15.0), np.radians(-20.0))):
        """
        Parameters
        ----------
        interp_points : `np.ndarray`, (N, 3)
            Interpolation points. Should contain sunAlt, airmass and azRelSun.
        max_am : `float`, optional
            Maximum airmass to calculate twilight sky to.
        limits : `np.ndarray`, (N,), optional
            Sun altitude limits

        Returns
        -------
        spectra, wavelength : `np.ndarray`, (N, 3), `np.ndarray`, (M,)

        Note
        ----
        Originally fit the twilight with a cutoff of sun altitude of
        -11 degrees. I think it can be safely extrapolated farther,
        but be warned you may be entering a regime where it breaks down.
        """

        npts = np.size(self.solar_wave)
        result = np.zeros((np.size(interp_points), npts), dtype=float)

        out_of_range = np.where(interp_points["sunAlt"] > np.radians(-11))[0]
        if np.size(out_of_range) > 0:
            warnings.warn("Extrapolating twilight beyond a sun altitude of -11 degrees")

        good = np.where(
            (interp_points["sunAlt"] >= np.min(limits))
            & (interp_points["sunAlt"] <= np.max(limits))
            & (interp_points["airmass"] <= max_am)
            & (interp_points["airmass"] >= 1.0)
        )[0]

        # Compute the expected flux in each of the filters that
        # we have fits for
        fluxes = []
        for filter_name in self.filter_names:
            out_of_range = np.where(interp_points["sunAlt"] > np.max(limits))[0]
            if np.size(out_of_range) > 0:
                fluxes.append(np.nan)
            else:
                fluxes.append(twilight_func(interp_points[good], *self.fit_results[filter_name]))
        fluxes = np.array(fluxes)

        # ratio of model flux to raw solar flux:
        yvals = fluxes.T / (10.0 ** (-0.4 * (self.solar_mag - np.log10(3631.0))))

        # Find wavelengths bluer than cutoff
        blue_region = np.where(self.solar_wave < np.min(self.eff_wave))

        for i, yval in enumerate(yvals):
            interp_f = interp1d(self.eff_wave, yval, bounds_error=False, fill_value=yval[-1])
            ratio = interp_f(self.solar_wave)
            interp_blue = InterpolatedUnivariateSpline(self.eff_wave, yval, k=1)
            ratio[blue_region] = interp_blue(self.solar_wave[blue_region])
            result[good[i]] = self.solar_flux * ratio

        return {"spec": result, "wave": self.solar_wave}


class MoonInterp(BaseSingleInterp):
    """
    Read in the saved Lunar spectra and interpolate.
    """

    def __init__(
        self,
        comp_name="Moon",
        sorted_order=["moonSunSep", "moonAltitude", "hpid"],
        mags=False,
    ):
        super(MoonInterp, self).__init__(comp_name=comp_name, sorted_order=sorted_order, mags=mags)
        # Magic number from when the templates were generated
        self.nside = 4

    def _weighting(self, interp_points, values):
        """
        Weighting for the scattered moonlight.
        """

        result = np.zeros((interp_points.size, np.size(values[0])), dtype=float)

        # Check that moonAltitude is in range, otherwise return zero array
        if np.max(interp_points["moonAltitude"]) < np.min(self.dim_dict["moonAltitude"]):
            return result

        # Find the neighboring healpixels
        hpids, hweights = get_neighbours(
            self.nside, np.pi / 2.0 - interp_points["alt"], interp_points["azRelMoon"]
        )

        badhp = np.isin(hpids.ravel(), self.dim_dict["hpid"], invert=True).reshape(hpids.shape)
        hweights[badhp] = 0.0

        norm = np.sum(hweights, axis=0)
        good = np.where(norm != 0.0)[0]
        hweights[:, good] = hweights[:, good] / norm[good]

        # Find the neighboring moonAltitude points in the grid
        right_m_as, left_m_as, ma_right_w, ma_left_w = self._indx_and_weights(
            interp_points["moonAltitude"], self.dim_dict["moonAltitude"]
        )

        # Find the neighboring moonSunSep points in the grid
        right_mss, left_mss, mss_right_w, mss_left_w = self._indx_and_weights(
            interp_points["moonSunSep"], self.dim_dict["moonSunSep"]
        )

        nhpid = self.dim_dict["hpid"].size
        n_ma = self.dim_dict["moonAltitude"].size
        # Convert the hpid to an index.
        tmp = intid2id(hpids.ravel(), self.dim_dict["hpid"], np.arange(self.dim_dict["hpid"].size))
        hpindx = tmp.reshape(hpids.shape)
        # loop though the hweights and the moonAltitude weights

        for hpid, hweight in zip(hpindx, hweights):
            for maid, maW in zip([right_m_as, left_m_as], [ma_right_w, ma_left_w]):
                for mssid, mssW in zip([right_mss, left_mss], [mss_right_w, mss_left_w]):
                    weight = hweight * maW * mssW
                    result += weight[:, np.newaxis] * values[mssid * nhpid * n_ma + maid * nhpid + hpid]

        return result


class ZodiacalInterp(BaseSingleInterp):
    """
    Interpolate the zodiacal light based on the airmass
    and the healpix ID where the healpixels are in ecliptic
    coordinates, with the sun at ecliptic longitude zero
    """

    def __init__(self, comp_name="Zodiacal", sorted_order=["airmass", "hpid"], mags=False):
        super(ZodiacalInterp, self).__init__(comp_name=comp_name, sorted_order=sorted_order, mags=mags)
        self.nside = hp.npix2nside(
            np.size(np.where(self.spec["airmass"] == np.unique(self.spec["airmass"])[0])[0])
        )

    def _weighting(self, interp_points, values):
        """
        interp_points is a numpy array where interpolation is desired
        values are the model values.
        """
        result = np.zeros((interp_points.size, np.size(values[0])), dtype=float)

        in_range = np.where(
            (interp_points["airmass"] <= np.max(self.dim_dict["airmass"]))
            & (interp_points["airmass"] >= np.min(self.dim_dict["airmass"]))
        )
        use_points = interp_points[in_range]
        # Find the neighboring healpixels
        hpids, hweights = get_neighbours(
            self.nside,
            np.pi / 2.0 - use_points["altEclip"],
            use_points["azEclipRelSun"],
        )

        badhp = np.isin(hpids.ravel(), self.dim_dict["hpid"], invert=True).reshape(hpids.shape)
        hweights[badhp] = 0.0

        norm = np.sum(hweights, axis=0)
        good = np.where(norm != 0.0)[0]
        hweights[:, good] = hweights[:, good] / norm[good]

        am_right_index, am_left_index, am_right_w, am_left_w = self._indx_and_weights(
            use_points["airmass"], self.dim_dict["airmass"]
        )

        nhpid = self.dim_dict["hpid"].size
        # loop though the hweights and the airmass weights
        for hpid, hweight in zip(hpids, hweights):
            for am_index, amW in zip([am_right_index, am_left_index], [am_right_w, am_left_w]):
                weight = hweight * amW
                result[in_range] += weight[:, np.newaxis] * values[am_index * nhpid + hpid]

        return result



__all__ = (
    "TemplateExistsMetric",
    "UniformityMetric",
    "GeneralUniformityMetric",
    "RapidRevisitUniformityMetric",
    "RapidRevisitMetric",
    "NRevisitsMetric",
    "IntraNightGapsMetric",
    "InterNightGapsMetric",
    "VisitGapMetric",
)

import numpy as np

from .base_metric import BaseMetric


class FSMetric(BaseMetric):
    """Calculate the fS value (Nvisit-weighted delta(M5-M5srd))."""

    def __init__(self, filter_col="filter", metric_name="fS", **kwargs):
        self.filter_col = filter_col
        cols = [self.filter_col]
        super().__init__(cols=cols, metric_name=metric_name, units="fS", **kwargs)

    def run(self, data_slice, slice_point=None):
        # We could import this from the m5_flat_sed values,
        # but it makes sense to calculate the m5
        # directly from the throughputs. This is easy enough to do and
        # will allow variation of
        # the throughput curves and readnoise and visit length, etc.
        pass


class TemplateExistsMetric(BaseMetric):
    """Calculate the fraction of images with a previous template
    image of desired quality."""

    def __init__(
        self,
        seeing_col="seeingFwhmGeom",
        observation_start_mjd_col="observationStartMJD",
        metric_name="TemplateExistsMetric",
        **kwargs,
    ):
        cols = [seeing_col, observation_start_mjd_col]
        super(TemplateExistsMetric, self).__init__(
            col=cols, metric_name=metric_name, units="fraction", **kwargs
        )
        self.seeing_col = seeing_col
        self.observation_start_mjd_col = observation_start_mjd_col

    def run(self, data_slice, slice_point=None):
        # Check that data is sorted in observationStartMJD order
        data_slice.sort(order=self.observation_start_mjd_col)
        # Find the minimum seeing up to a given time
        seeing_mins = np.minimum.accumulate(data_slice[self.seeing_col])
        # Find the difference between the seeing and the minimum seeing
        # at the previous visit
        seeing_diff = data_slice[self.seeing_col] - np.roll(seeing_mins, 1)
        # First image never has a template; check how many others do
        good = np.where(seeing_diff[1:] >= 0.0)[0]
        frac = (good.size) / float(data_slice[self.seeing_col].size)
        return frac


class UniformityMetric(BaseMetric):
    """Calculate how uniformly the observations are spaced in time.

    This is based on how a KS-test works:
    look at the cumulative distribution of observation dates,
    and compare to a perfectly uniform cumulative distribution.
    Perfectly uniform observations = 0, perfectly non-uniform = 1.

    Parameters
    ----------
    mjd_col : `str`, optional
        The column containing time for each observation.
        Default "observationStartMJD".
    survey_length : `float`, optional
        The overall duration of the survey. Default 10.
    """

    def __init__(self, mjd_col="observationStartMJD", units="", survey_length=10.0, **kwargs):
        """survey_length = time span of survey (years)"""
        self.mjd_col = mjd_col
        super(UniformityMetric, self).__init__(col=self.mjd_col, units=units, **kwargs)
        self.survey_length = survey_length

    def run(self, data_slice, slice_point=None):
        # If only one observation, there is no uniformity
        if data_slice[self.mjd_col].size == 1:
            return 1
        # Scale dates to lie between 0 and 1,
        # where 0 is the first observation date and 1 is surveyLength
        dates = (data_slice[self.mjd_col] - data_slice[self.mjd_col].min()) / (self.survey_length * 365.25)
        dates.sort()  # Just to be sure
        n_cum = np.arange(1, dates.size + 1) / float(dates.size)
        d_max = np.max(np.abs(n_cum - dates - dates[1]))
        return d_max


class GeneralUniformityMetric(BaseMetric):
    """Calculate how uniformly any values are distributed.

    This is based on how a KS-test works:
    look at the cumulative distribution of data,
    and compare to a perfectly uniform cumulative distribution.
    Perfectly uniform observations = 0, perfectly non-uniform = 1.
    To be "perfectly uniform" here, the endpoints need to be included.

    Parameters
    ----------
    col : `str`, optional
        The column of data to use for the metric.
        The default is "observationStartMJD" as this is most
        typically used with time.
    min_value : `float`, optional
        The minimum value expected for the data.
        Default None will calculate use the minimum value in this dataslice
        (which may not cover the full range).
    max_value : `float`, optional
        The maximum value expected for the data.
        Default None will calculate use the maximum value in this dataslice
        (which may not cover the full range).
    """

    def __init__(self, col="observationStartMJD", units="", min_value=None, max_value=None, **kwargs):
        self.col = col
        super().__init__(col=self.col, units=units, **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def run(self, data_slice, slice_point=None):
        # If only one observation, there is no uniformity
        if data_slice[self.col].size == 1:
            return 1
        # Scale values to lie between 0 and 1,
        # where 0 is the min_value and 1 is max_value
        if self.min_value is None:
            min_value = data_slice[self.col].min()
        else:
            min_value = self.min_value
        if self.max_value is None:
            max_value = data_slice[self.col].max()
        else:
            max_value = self.max_value
        scaled_values = (data_slice[self.col] - min_value) / max_value
        scaled_values.sort()  # Just to be sure
        n_cum = np.arange(0, scaled_values.size) / float(scaled_values.size - 1)
        d_max = np.max(np.abs(n_cum - scaled_values))
        return d_max


class RapidRevisitUniformityMetric(BaseMetric):
    """Calculate uniformity of time between consecutive visits on
    short timescales (for RAV1).

    Uses the same 'uniformity' calculation as the UniformityMetric,
    based on the KS-test.
    A value of 0 is perfectly uniform; a value of 1 is purely non-uniform.

    Parameters
    ----------
    mjd_col : `str`, optional
        The column containing the 'time' value. Default observationStartMJD.
    min_nvisits : `int`, optional
        The minimum number of visits required within the
        time interval (d_tmin to d_tmax).
        Default 100.
    d_tmin : `float`, optional
        The minimum dTime to consider (in days). Default 40 seconds.
    d_tmax : `float`, optional
        The maximum dTime to consider (in days). Default 30 minutes.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        min_nvisits=100,
        d_tmin=40.0 / 60.0 / 60.0 / 24.0,
        d_tmax=30.0 / 60.0 / 24.0,
        metric_name="RapidRevisitUniformity",
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.min_nvisits = min_nvisits
        self.d_tmin = d_tmin
        self.d_tmax = d_tmax
        super().__init__(col=self.mjd_col, metric_name=metric_name, **kwargs)
        # Update min_nvisits, as 0 visits will crash algorithm
        # and 1 is nonuniform by definition.
        if self.min_nvisits <= 1:
            self.min_nvisits = 2

    def run(self, data_slice, slice_point=None):
        # Calculate consecutive visit time intervals
        dtimes = np.diff(np.sort(data_slice[self.mjd_col]))
        # Identify dtimes within interval from dTmin/dTmax.
        good = np.where((dtimes >= self.d_tmin) & (dtimes <= self.d_tmax))[0]
        # If there are not enough visits in this time range, return bad value.
        if good.size < self.min_nvisits:
            return self.badval
        # Throw out dtimes outside desired range, and sort, then scale to 0-1.
        dtimes = np.sort(dtimes[good])
        dtimes = (dtimes - dtimes.min()) / float(self.d_tmax - self.d_tmin)
        # Set up a uniform distribution between 0-1 (to match dtimes).
        uniform_dtimes = np.arange(1, dtimes.size + 1, 1) / float(dtimes.size)
        # Look at the differences between our times and the uniform times.
        dmax = np.max(np.abs(uniform_dtimes - dtimes - dtimes[1]))
        return dmax


class RapidRevisitMetric(BaseMetric):
    def __init__(
        self,
        mjd_col="observationStartMJD",
        metric_name="RapidRevisit",
        d_tmin=40.0 / 60.0 / 60.0 / 24.0,
        d_tpairs=20.0 / 60.0 / 24.0,
        d_tmax=30.0 / 60.0 / 24.0,
        min_n1=28,
        min_n2=82,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.d_tmin = d_tmin
        self.d_tpairs = d_tpairs
        self.d_tmax = d_tmax
        self.min_n1 = min_n1
        self.min_n2 = min_n2
        super().__init__(col=self.mjd_col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        dtimes = np.diff(np.sort(data_slice[self.mjd_col]))
        n1 = len(np.where((dtimes >= self.d_tmin) & (dtimes <= self.d_tpairs))[0])
        n2 = len(np.where((dtimes >= self.d_tmin) & (dtimes <= self.d_tmax))[0])
        if (n1 >= self.min_n1) and (n2 >= self.min_n2):
            val = 1
        else:
            val = 0
        return val


class NRevisitsMetric(BaseMetric):
    """Calculate the number of consecutive visits with
    time differences less than d_t.

    Parameters
    ----------
    d_t : `float`, optional
        The time interval to consider (in minutes). Default 30.
    normed : `bool`, optional
        Flag to indicate whether to return the total number of
        consecutive visits with time differences less than d_t (False),
        or the fraction of overall visits (True).
        Note that we would expect (if all visits occur in pairs within d_t)
        this fraction would be 0.5!
    """

    def __init__(self, mjd_col="observationStartMJD", d_t=30.0, normed=False, metric_name=None, **kwargs):
        units = ""
        if metric_name is None:
            if normed:
                metric_name = "Fraction of revisits faster than %.1f minutes" % (d_t)
            else:
                metric_name = "Number of revisits faster than %.1f minutes" % (d_t)
                units = "#"
        self.mjd_col = mjd_col
        self.d_t = d_t / 60.0 / 24.0  # convert to days
        self.normed = normed
        super(NRevisitsMetric, self).__init__(
            col=self.mjd_col, units=units, metric_name=metric_name, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        dtimes = np.diff(np.sort(data_slice[self.mjd_col]))
        n_fast_revisits = np.size(np.where(dtimes <= self.d_t)[0])
        if self.normed:
            n_fast_revisits = n_fast_revisits / float(np.size(data_slice[self.mjd_col]))
        return n_fast_revisits


class IntraNightGapsMetric(BaseMetric):
    """
    Calculate the (reduce_func) of the gap between consecutive
    observations within a night, in hours.

    Parameters
    ----------
    reduce_func : function, optional
        Function that can operate on array-like structures.
        Typically numpy function.
        Default np.median.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        night_col="night",
        reduce_func=np.median,
        metric_name="Median Intra-Night Gap",
        **kwargs,
    ):
        units = "hours"
        self.mjd_col = mjd_col
        self.night_col = night_col
        self.reduce_func = reduce_func
        super(IntraNightGapsMetric, self).__init__(
            col=[self.mjd_col, self.night_col], units=units, metric_name=metric_name, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.mjd_col)
        dt = np.diff(data_slice[self.mjd_col])
        dn = np.diff(data_slice[self.night_col])

        good = np.where(dn == 0)
        if np.size(good[0]) == 0:
            result = self.badval
        else:
            result = self.reduce_func(dt[good]) * 24
        return result


class InterNightGapsMetric(BaseMetric):
    """Calculate the (reduce_func) of the gap between consecutive
    observations in different nights, in days.

    Parameters
    ----------
    reduce_func : function, optional
        Function that can operate on array-like structures.
        Typically numpy function.
        Default np.median.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        night_col="night",
        reduce_func=np.median,
        metric_name="Median Inter-Night Gap",
        **kwargs,
    ):
        units = "days"
        self.mjd_col = mjd_col
        self.night_col = night_col
        self.reduce_func = reduce_func
        super().__init__(col=[self.mjd_col, self.night_col], units=units, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.mjd_col)
        unights = np.unique(data_slice[self.night_col])
        if np.size(unights) < 2:
            result = self.badval
        else:
            # Find the first and last observation of each night
            first_of_night = np.searchsorted(data_slice[self.night_col], unights)
            last_of_night = np.searchsorted(data_slice[self.night_col], unights, side="right") - 1
            diff = data_slice[self.mjd_col][first_of_night[1:]] - data_slice[self.mjd_col][last_of_night[:-1]]
            result = self.reduce_func(diff)
        return result


class VisitGapMetric(BaseMetric):
    """Calculate the (reduce_func) of the gap between any
    consecutive observations, in hours, regardless of night boundaries.

    Different from inter-night and intra-night gaps,
    because this is really just counting all of the times between consecutive
    observations (not time between nights or time within a night).

    Parameters
    ----------
    reduce_func : function, optional
        Function that can operate on array-like structures.
        Typically numpy function.
        Default np.median.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        night_col="night",
        reduce_func=np.median,
        metric_name="VisitGap",
        **kwargs,
    ):
        units = "hours"
        self.mjd_col = mjd_col
        self.night_col = night_col
        self.reduce_func = reduce_func
        super().__init__(col=[self.mjd_col, self.night_col], units=units, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.mjd_col)
        diff = np.diff(data_slice[self.mjd_col])
        result = self.reduce_func(diff) * 24.0
        return result



import os
import unittest
import warnings

import numpy as np
from rubin_scheduler.data import get_data_dir

import rubin_sim.maf.maps as maps
import rubin_sim.maf.slicers as slicers


def make_data_values(size=100, min=0.0, max=1.0, random=-1):
    """Generate a simple array of numbers, evenly arranged between min/max,
    (optional) random order.
    """
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(max) - float(min)) / (datavalues.max() - datavalues.min())
    datavalues += min
    rot_vals = datavalues * 0
    if random > 0:
        rng = np.random.RandomState(random)
        randorder = rng.rand(size)
        randind = np.argsort(randorder)
        datavalues = datavalues[randind]
    ids = np.arange(size)
    datavalues = np.array(
        list(zip(datavalues, datavalues, ids, rot_vals)),
        dtype=[
            ("fieldRA", "float"),
            ("fieldDec", "float"),
            ("fieldId", "int"),
            ("rotSkyPos", "float"),
        ],
    )
    return datavalues


def make_field_data(seed):
    rng = np.random.RandomState(seed)
    names = ["fieldId", "fieldRA", "fieldDec"]
    types = [int, float, float]
    field_data = np.zeros(100, dtype=list(zip(names, types)))
    field_data["fieldId"] = np.arange(100)
    field_data["fieldRA"] = rng.rand(100)
    field_data["fieldDec"] = rng.rand(100)
    return field_data


class TestMaps(unittest.TestCase):
    def test_dust_map(self):
        map_path = os.path.join(get_data_dir(), "tests")
        nside = 8
        if os.path.isfile(os.path.join(map_path, f"dust_nside_{nside}.npz")):
            data = make_data_values(random=981)
            dustmap = maps.DustMap(nside=nside, map_path=map_path)

            slicer1 = slicers.HealpixSlicer(lat_lon_deg=False, nside=nside, use_camera=False)
            slicer1.setup_slicer(data)
            result1 = dustmap.run(slicer1.slice_points)
            assert "ebv" in list(result1.keys())

            field_data = make_field_data(2234)

            slicer2 = slicers.UserPointsSlicer(
                field_data["fieldRA"], field_data["fieldDec"], lat_lon_deg=False
            )
            result2 = dustmap.run(slicer2.slice_points)
            assert "ebv" in list(result2.keys())

            # Check interpolation works
            dustmap = maps.DustMap(interp=True, nside=nside, map_path=map_path)
            result3 = dustmap.run(slicer2.slice_points)
            assert "ebv" in list(result3.keys())

            # Check warning gets raised
            dustmap = maps.DustMap(nside=4, map_path=map_path)
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                dustmap.run(slicer1.slice_points)
                self.assertIn("nside", str(w[-1].message))
        else:
            warnings.warn("Did not find dustmaps, not running testMaps.py")

    def test_dust_map3_d(self):
        nside = 8
        map_file = os.path.join(get_data_dir(), "tests", f"test_ebv3d_nside{nside}.fits")
        if os.path.isfile(map_file):
            data = make_data_values(random=981)
            dustmap = maps.DustMap3D(nside=nside, map_file=map_file, interp=False)

            slicer1 = slicers.HealpixSlicer(lat_lon_deg=False, nside=nside, use_camera=False)
            slicer1.setup_slicer(data)
            result1 = dustmap.run(slicer1.slice_points)
            assert "ebv3d_ebvs" in list(result1.keys())
            assert "ebv3d_dists" in list(result1.keys())

            field_data = make_field_data(2234)

            slicer2 = slicers.UserPointsSlicer(
                field_data["fieldRA"], field_data["fieldDec"], lat_lon_deg=False
            )
            result2 = dustmap.run(slicer2.slice_points)
            assert "ebv3d_ebvs" in list(result2.keys())
            assert "ebv3d_dists" in list(result2.keys())

            # Check interpolation works
            dustmap = maps.DustMap3D(interp=True, nside=nside, map_file=map_file, dist_pc=2000, d_mag=10)
            result3 = dustmap.run(slicer2.slice_points)
            assert "ebv3d_ebvs" in list(result3.keys())
            assert "ebv3d_dists" in list(result3.keys())
            assert "ebv3d_ebv_at_2000.0" in list(result3.keys())
            assert "ebv3d_dist_at_10.0" in list(result3.keys())

            # Check that we can call the distance at magnitude method
            _ = dustmap.distance_at_dmag(5, result3["ebv3d_dists"], result3["ebv3d_ebvs"], "r")
            # And call it without running the map
            # (and thus reading more info) first
            _ = maps.DustMap3D().distance_at_dmag(5, result3["ebv3d_dists"], result3["ebv3d_ebvs"], "r")
            # And call it as a static method at one point on the sky
            _ = maps.DustMap3D().distance_at_dmag(
                5, result3["ebv3d_dists"][0, :], result3["ebv3d_ebvs"][0, :], "r"
            )

            # Check warning gets raised
            dustmap = maps.DustMap3D(nside=4, map_file=map_file)
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                dustmap.run(slicer1.slice_points)
                self.assertIn("nside", str(w[-1].message))
        else:
            warnings.warn("Did not find dustmaps, not running testMaps.py")

    def test_star_map(self):
        map_path = os.path.join(get_data_dir(), "tests")

        if os.path.isfile(os.path.join(map_path, "starDensity_r_nside_64.npz")):
            data = make_data_values(random=887)
            # check that it works if nside does not match map nside of 64
            nsides = [32, 64, 128]
            for nside in nsides:
                starmap = maps.StellarDensityMap(map_dir=map_path)
                slicer1 = slicers.HealpixSlicer(nside=nside, lat_lon_deg=False, use_camera=False)
                slicer1.setup_slicer(data)
                result1 = starmap.run(slicer1.slice_points)
                assert "starMapBins_r" in list(result1.keys())
                assert "starLumFunc_r" in list(result1.keys())
                assert np.max(result1["starLumFunc_r"] > 0)

            field_data = make_field_data(22)

            slicer2 = slicers.UserPointsSlicer(
                field_data["fieldRA"], field_data["fieldDec"], lat_lon_deg=False
            )
            result2 = starmap.run(slicer2.slice_points)
            assert "starMapBins_r" in list(result2.keys())
            assert "starLumFunc_r" in list(result2.keys())
            assert np.max(result2["starLumFunc_r"] > 0)

        else:
            warnings.warn("Did not find stellar density map, skipping test.")

    @unittest.skipUnless(
        os.path.isdir(os.path.join(get_data_dir(), "maps")),
        "Skip the galplane priority map data unless maps data present, required for setup",
    )
    def test_galplane_priority_maps(self):
        map_path = os.path.join(get_data_dir(), "maps")
        nside = 64
        data = make_data_values(random=981)
        galplane_map = maps.GalacticPlanePriorityMap(nside=nside, map_path=map_path, interp=False)

        # Set up basic case - healpix
        slicer1 = slicers.HealpixSlicer(lat_lon_deg=False, nside=nside, use_camera=False)
        slicer1.setup_slicer(data)
        result1 = galplane_map.run(slicer1.slice_points)
        key = maps.gp_priority_map_components_to_keys("sum", "combined_map")
        assert key in list(result1.keys())

        # Set up more advanced case - random ra/dec
        field_data = make_field_data(2234)
        slicer2 = slicers.UserPointsSlicer(field_data["fieldRA"], field_data["fieldDec"], lat_lon_deg=False)
        result2 = galplane_map.run(slicer2.slice_points)
        assert key in list(result2.keys())

        # Check interpolation works
        galplane_map = maps.GalacticPlanePriorityMap(interp=True, nside=nside, map_path=map_path)
        result3 = galplane_map.run(slicer2.slice_points)
        assert key in list(result3.keys())

        # Check warning gets raised
        galplane_map = maps.GalacticPlanePriorityMap(nside=4, map_path=map_path)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            galplane_map.run(slicer1.slice_points)
            self.assertIn("nside", str(w[-1].message))


if __name__ == "__main__":
    unittest.main()



__all__ = ("sigma_slope", "m52snr", "astrom_precision")

import numpy as np

"""Some simple functions that are useful for astrometry calculations. """


def sigma_slope(x, sigma_y):
    """
    Calculate the uncertainty in fitting a line, as
    given by the spread in x values and the uncertainties
    in the y values.

    Parameters
    ----------
    x : numpy.ndarray
        The x values of the data
    sigma_y : numpy.ndarray
        The uncertainty in the y values

    Returns
    -------
    float
        The uncertainty in the line fit
    """
    w = 1.0 / sigma_y**2
    denom = np.sum(w) * np.sum(w * x**2) - np.sum(w * x) ** 2
    if denom <= 0:
        return np.nan
    else:
        result = np.sqrt(np.sum(w) / denom)
        return result


def m52snr(m, m5, gamma=0.04):
    """
    Calculate the SNR for a star of magnitude m in an
    observation with 5-sigma limiting magnitude depth m5.
    Assumes gaussian distribution of photons and might not be
    strictly due in bluer filters. See table 2 and equation 5
    in astroph/0805.2366.

    Parameters
    ----------
    m : `float` or `np.ndarray` (N,)
        The magnitude of the star
    m5 : `float` or `np.ndarray` (N,)
        The m5 limiting magnitude of the observation
    gamma : `float` or None
        The 'gamma' value used when calculating photometric or
        astrometric errors and weighting SNR accordingly.
        See equation 5 of the LSST Overview paper.
        Use "None" to discount the gamma factor completely
        and use standard 5*10^(0.4 * (m5-m)).

    Returns
    -------
    snr : `float` or `np.ndarray` (N,)
        The SNR
    """
    # gamma varies per band, but is fairly close to 0.04

    if gamma is None:
        snr = 5.0 * 10.0 ** (-0.4 * (m - m5))
    else:
        xval = np.power(10, 0.4 * (m - m5))
        snr = 1 / np.sqrt((0.04 - gamma) * xval + gamma * xval * xval)
    return snr


def astrom_precision(fwhm, snr, systematic_floor=0.00):
    """
    Calculate the approximate precision of astrometric measurements,
    given a particular seeing and SNR value.

    Parameters
    ----------
    fwhm : `float` or `np.ndarray` (N,)
        The seeing (FWHMgeom) of the observation.
    snr : float` or `np.ndarray` (N,)
        The SNR of the object.
    systematic_floor : `float`
        Systematic noise floor for astrometric error, in arcseconds.
        Default here is 0, for backwards compatibility.
        General Rubin use should be 0.01.

    Returns
    -------
    astrom_err : `float` or `numpy.ndarray` (N,)
        The astrometric precision, in arcseconds.
    """
    astrom_err = np.sqrt((fwhm / snr) ** 2 + systematic_floor**2)
    return astrom_err



import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
import pandas as pd
from astropy.time import Time
from rubin_scheduler.data import get_data_dir

from rubin_sim.moving_objects import ChebyFits, ChebyValues, Orbits, PyOrbEphemerides

ROOT = os.path.abspath(os.path.dirname(__file__))


@unittest.skip("Temporary skip until ephemerides replaced")
class TestChebyValues(unittest.TestCase):
    def setUp(self):
        self.testdatadir = os.path.join(get_data_dir(), "tests", "orbits_testdata")
        self.scratch_dir = tempfile.mkdtemp(dir=ROOT, prefix="TestChebyValues-")
        self.coeff_file = os.path.join(self.scratch_dir, "test_coeffs")
        self.resid_file = os.path.join(self.scratch_dir, "test_resids")
        self.failed_file = os.path.join(self.scratch_dir, "test_failed")
        self.orbits = Orbits()
        self.orbits.read_orbits(os.path.join(self.testdatadir, "test_orbitsNEO.s3m"), skiprows=1)
        self.pyephems = PyOrbEphemerides()
        self.pyephems.set_orbits(self.orbits)
        self.t_start = self.orbits.orbits.epoch.iloc[0]
        self.interval = 30
        self.n_coeffs = 14
        self.n_decimal = 13

        self.cheby_fits = ChebyFits(
            self.orbits,
            self.t_start,
            self.interval,
            ngran=64,
            sky_tolerance=2.5,
            n_decimal=self.n_decimal,
            n_coeff_position=self.n_coeffs,
            obscode=807,
            time_scale="TAI",
        )
        self.set_length = 0.5
        self.cheby_fits.calc_segment_length(length=self.set_length)
        self.cheby_fits.calc_segments()
        self.cheby_fits.write(self.coeff_file, self.resid_file, self.failed_file, append=False)
        self.coeff_keys = [
            "obj_id",
            "t_start",
            "t_end",
            "ra",
            "dec",
            "geo_dist",
            "vmag",
            "elongation",
        ]

    def tearDown(self):
        del self.orbits
        del self.cheby_fits
        if os.path.exists(self.scratch_dir):
            shutil.rmtree(self.scratch_dir)

    def test_set_coeff(self):
        # Test setting coefficients directly from chebyFits outputs.
        cheby_values = ChebyValues()
        cheby_values.set_coefficients(self.cheby_fits)
        for k in self.coeff_keys:
            self.assertTrue(k in cheby_values.coeffs)
            self.assertTrue(isinstance(cheby_values.coeffs[k], np.ndarray))
        self.assertEqual(len(np.unique(cheby_values.coeffs["obj_id"])), len(self.orbits))
        # This will only be true for carefully selected length/orbit type,
        # where subdivision did not occur.
        # For the test MBAs, a len=1day will work.
        # For the test NEOs, a len=0.25 day will work (with 2.5mas skyTol).
        # self.assertEqual(len(cheby_values.coeffs['tStart']),
        #            (self.interval / self.set_length) * len(self.orbits))
        self.assertEqual(len(cheby_values.coeffs["ra"][0]), self.n_coeffs)
        self.assertTrue("meanRA" in cheby_values.coeffs)
        self.assertTrue("meanDec" in cheby_values.coeffs)

    def test_read_coeffs(self):
        # Test reading the coefficients from disk.
        cheby_values = ChebyValues()
        cheby_values.read_coefficients(self.coeff_file)
        cheby_values2 = ChebyValues()
        cheby_values2.set_coefficients(self.cheby_fits)
        for k in cheby_values.coeffs:
            if k == "obj_id":
                # Can't test strings with np.test.assert_almost_equal.
                np.testing.assert_equal(cheby_values.coeffs[k], cheby_values2.coeffs[k])
            else:
                # All of these will only be accurate to 2 less decimal places
                # than they are print out with in chebyFits. Since vmag,
                # delta and elongation only use 7
                # decimal places, this means we can test to 5 decimal
                # places for those.
                np.testing.assert_allclose(cheby_values.coeffs[k], cheby_values2.coeffs[k], rtol=0, atol=1e-5)

    def test_get_ephemerides(self):
        # Test that get_ephemerides works and is accurate.
        cheby_values = ChebyValues()
        cheby_values.set_coefficients(self.cheby_fits)

        # Multiple times, all objects, all within interval.
        tstep = self.interval / 10.0
        time = np.arange(self.t_start, self.t_start + self.interval, tstep)
        # Test for a single time, but all the objects.
        ephemerides = cheby_values.get_ephemerides(time)
        pyephemerides = self.pyephems.generate_ephemerides(
            time, obscode=807, time_scale="TAI", by_object=True
        )
        # Looks like there's some weird hardware dependent issue
        # where cheby_fits is failing to fit some orbits.

        # RA and Dec should agree to 2.5mas (sky_tolerance above)
        pos_residuals = np.sqrt(
            (ephemerides["ra"] - pyephemerides["ra"]) ** 2
            + ((ephemerides["dec"] - pyephemerides["dec"]) * np.cos(np.radians(ephemerides["dec"]))) ** 2
        )
        pos_residuals *= 3600.0 * 1000.0
        if not np.all(np.isfinite(ephemerides["ra"])):
            warnings.warn("NaN values from ChebyValues.get_ephemerides, skipping some tests")
        else:
            # Let's just look at the max residuals in all quantities.
            for k in ("ra", "dec", "dradt", "ddecdt", "geo_dist"):
                resids = np.abs(ephemerides[k] - pyephemerides[k])
                if k != "geo_dist":
                    resids *= 3600.0 * 1000.0
                print("max diff", k, np.max(resids))
            resids = np.abs(ephemerides["elongation"] - pyephemerides["solarelon"])
            print("max diff elongation", np.max(resids))
            resids = np.abs(ephemerides["vmag"] - pyephemerides["magV"])
            print("max diff vmag", np.max(resids))
            self.assertLessEqual(np.nanmax(pos_residuals), 2.5)
            # Test for single time, but for a subset of the objects.
            obj_ids = self.orbits.orbits.obj_id.head(3).values
            ephemerides = cheby_values.get_ephemerides(time, obj_ids)
            self.assertEqual(len(ephemerides["ra"]), 3)
            # Test for time outside of segment range.
            obj_ids = self.orbits.orbits.obj_id.head(3).values
            ephemerides = cheby_values.get_ephemerides(
                self.t_start + self.interval * 2, obj_ids, extrapolate=False
            )
            self.assertTrue(
                np.isnan(ephemerides["ra"][0]),
                msg=f"Expected Nan for out of range ephemeris, got {ephemerides['ra'][0]}",
            )


@unittest.skip("Temporary skip until ephemerides replaced")
class TestJPLValues(unittest.TestCase):
    # Test the interpolation-generated RA/Dec values against JPL
    # generated RA/Dec values.
    # The resulting errors should be similar to the errors reported
    # from testEphemerides when testing against JPL values.
    def setUp(self):
        # Read orbits.
        self.orbits = Orbits()
        self.jpl_dir = os.path.join(get_data_dir(), "tests", "jpl_testdata")
        self.orbits.read_orbits(os.path.join(self.jpl_dir, "S0_n747.des"), skiprows=1)
        # Read JPL ephems.
        self.jpl = pd.read_table(os.path.join(self.jpl_dir, "807_n747.txt"), sep=r"\s+")
        self.jpl["obj_id"] = self.jpl["objId"]
        # Add times in TAI and UTC, because.
        t = Time(self.jpl["epoch_mjd"], format="mjd", scale="utc")
        self.jpl["mjdTAI"] = t.tai.mjd
        self.jpl["mjdUTC"] = t.utc.mjd
        self.jpl = self.jpl.to_records(index=False)
        # Generate interpolation coefficients for the time period
        # in the JPL catalog.
        self.scratch_dir = tempfile.mkdtemp(dir=ROOT, prefix="TestJPLValues-")
        self.coeff_file = os.path.join(self.scratch_dir, "test_coeffs")
        self.resid_file = os.path.join(self.scratch_dir, "test_resids")
        self.failed_file = os.path.join(self.scratch_dir, "test_failed")
        t_start = self.jpl["mjdTAI"].min() - 0.2
        t_end = self.jpl["mjdTAI"].max() + 0.2 - self.jpl["mjdTAI"].min()
        self.cheby_fits = ChebyFits(
            self.orbits,
            t_start,
            t_end,
            ngran=64,
            sky_tolerance=2.5,
            n_decimal=14,
            obscode=807,
        )
        self.cheby_fits.calc_segment_length()
        self.cheby_fits.calc_segments()

        self.coeff_keys = [
            "obj_id",
            "t_start",
            "t_end",
            "ra",
            "dec",
            "geo_dist",
            "vmag",
            "elongation",
        ]
        self.cheby_values = ChebyValues()
        self.cheby_values.set_coefficients(self.cheby_fits)

    def tearDown(self):
        del self.orbits
        del self.jpl
        if os.path.exists(self.scratch_dir):
            shutil.rmtree(self.scratch_dir)

    def test_ra_dec(self):
        # We won't compare Vmag, because this also needs
        # information on trailing losses.
        times = np.unique(self.jpl["mjdTAI"])
        delta_ra = np.zeros(len(times), float)
        delta_dec = np.zeros(len(times), float)
        for i, t in enumerate(times):
            # Find the JPL obj_ids visible at this time.
            j = self.jpl[np.where(self.jpl["mjdTAI"] == t)]
            ephs = self.cheby_values.get_ephemerides(t, obj_ids=j["obj_id"])
            ephorder = np.argsort(ephs["obj_id"])
            # Sometimes I've had to reorder both, sometimes just the ephs. ??
            jorder = np.argsort(j["obj_id"])
            jorder = np.arange(0, len(jorder))
            d_ra = np.abs(ephs["ra"][ephorder][:, 0] - j["ra_deg"][jorder]) * 3600.0 * 1000.0
            d_dec = np.abs(ephs["dec"][ephorder][:, 0] - j["dec_deg"][jorder]) * 3600.0 * 1000.0
            delta_ra[i] = d_ra.max()
            delta_dec[i] = d_dec.max()
        # Should be (given OOrb direct prediction):
        # Much of the time we're closer than 1mas, but there are a
        # few which hit higher values.
        # This is consistent with the errors/values reported by oorb
        # directly in testEphemerides.

        #    # XXX--units?
        print("max JPL errors", delta_ra.max(), delta_dec.max())
        print("std of JPL errors", np.std(delta_ra), np.std(delta_dec))
        self.assertLess(np.max(delta_ra), 25)
        self.assertLess(np.max(delta_dec), 25)
        self.assertLess(np.std(delta_ra), 3)
        self.assertLess(np.std(delta_dec), 3)


if __name__ == "__main__":
    unittest.main()



"""Utilities to estimate and evaluate Chebyshev coefficients of a function.

Implementation of Newhall, X. X. 1989, Celestial Mechanics, 45, p. 305-310
"""

__all__ = ("chebeval", "chebfit", "make_cheb_matrix", "make_cheb_matrix_only_x")

import numpy as np

# Evaluation routine.


def chebeval(x, p, interval=(-1.0, 1.0), do_velocity=True, mask=False):
    """Evaluate a Chebyshev series and first derivative at points x.

    If p is of length n + 1, this function returns:
    y_hat(x) = p_0 * T_0(x*) + p_1 * T_1(x*) + ... + p_n * T_n(x*)
    where T_n(x*) are the orthogonal Chebyshev polynomials of the
    first kind, defined on the interval [-1, 1] and p_n are the
    coefficients. The scaled variable x* is defined on the [-1, 1]
    interval such that (x*) = (2*x - a - b)/(b - a), and x is defined
    on the [a, b] interval.

    Parameters
    ----------
    x : `scalar` or `np.ndarray`
        Points at which to evaluate the polynomial.
    p :  `np.ndarray`
        Chebyshev polynomial coefficients, as returned by chebfit.
    interval : 2-element list/tuple
        Bounds the x-interval on which the Chebyshev coefficients were fit.
    do_velocity : `bool`
        If True, compute the first derivative at points x.
    mask : `bool`
        If True, return Nans when the x goes beyond 'interval'.
        If False, extrapolate fit beyond 'interval' limits.

    Returns
    -------
    y, v : `float` or `np.ndarray`, `float` or `np.ndarray` (or None)
        Y (position) and velocity values (if computed)
    """
    if len(interval) != 2:
        raise RuntimeError("interval must have length 2")

    interval_begin = float(interval[0])
    interval_end = float(interval[-1])
    t = 2.0 * np.array(x, dtype=np.float64) - interval_begin - interval_end
    t /= interval_end - interval_begin

    y = 0.0
    v = 0.0
    y0 = np.ones_like(t)
    y1 = t
    v0 = np.zeros_like(t)
    v1 = np.ones_like(t)
    v2 = 4.0 * t
    t = 2.0 * t
    N = len(p)

    if do_velocity:
        for i in np.arange(0, N, 2):
            if i == N - 1:
                y1 = 0.0
                v1 = 0.0
            j = min(i + 1, N - 1)

            y += p[i] * y0 + p[j] * y1
            v += p[i] * v0 + p[j] * v1

            y2 = t * y1 - y0
            y3 = t * y2 - y1
            v2 = t * v1 - v0 + 2 * y1
            v3 = t * v2 - v1 + 2 * y2

            y0 = y2
            y1 = y3
            v0 = v2
            v1 = v3

        if mask:
            mask = np.where((x < interval_begin) | (x > interval_end), True, False)
            y = np.where(mask, np.nan, y)
            v = np.where(mask, np.nan, v)
        return y, 2 * v / (interval_end - interval_begin)
    else:
        for i in np.arange(0, N, 2):
            if i == N - 1:
                y1 = 0.0
            j = min((i + 1), (N - 1))
            y += p[i] * y0 + p[j] * y1
            y0 = t * y1 - y0
            y1 = t * y0 - y1
        if mask:
            mask = np.where((x < interval_begin) | (x > interval_end), True, False)
            y = np.where(mask, np.nan, y)
        return y, None


# Fitting routines.


def make_cheb_matrix(n_points, n_poly, weight=0.16):
    """Compute C1^(-1)C2 using Newhall89 approach.

    Utility function for fitting chebyshev polynomials to
    x(t) and dx/dt(t) forcing equality at the end points.
    This function computes the matrix (C1^(-1)C2).
    Multiplying this matrix by the x and dx/dt values to be fit
    produces the chebyshev coefficient.
    This function need only be called once for a given polynomial degree and
    number of points.

    The matrices returned are of shape(n_points+1)x(n_poly).
    The coefficients fitting the n_points+1 points, X, are found by:
    A = xMultiplier * x  +  dxMultiplier * dxdt
    if derivative information is known, or
    A = xMultiplier * x
    if no derivative information is known.
    The xMultiplier matrices are different,
    depending on whether derivative information is known.
    Use function make_cheb_matrix_only_x if derviative is not known.
    See Newhall, X. X. 1989, Celestial Mechanics, 45, p. 305-310 for details.

    Parameters
    ----------
    n_points : `int`
        Number of point to be fits. Must be greater than 2.
    n_poly :  `int`
        Number of polynomial terms. Polynomial degree + 1
    weight : `float`, optional
        Weight to allow control of relative effectos of position and velocity
        values. Newhall80 found best results are obtained with
        velocity weighted at 0.4 relative to position,
        giving W the form (1.0, 0.16, 1.0, 0.16,...)

    Returns
    -------
    c1c2: `np.ndarray`
        xMultiplier, C1^(-1)C2 even rows of shape (n_points+1)x(n_poly) to
        be multiplied by x values.
    c1c2: `np.ndarray`
        dxMultiplier, C1^(-1)C2 odd rows of shape (n_points+1)x(n_poly) to
        be multiplied by dx/dy values
    """
    tmat = np.zeros([n_points, n_poly])
    tdot = np.zeros([n_points, n_poly])

    cj = np.zeros([n_poly])
    xj = np.linspace(1, -1, n_points)

    for i in np.arange(0, n_poly):
        cj[:] = 0
        cj[i] = 1
        y, v = chebeval(xj, cj)
        tmat[:, i] = y
        tdot[:, i] = v

    # make matrix T*W
    tw = np.zeros([n_poly, n_points, 2])
    tw[:, :, 0] = tmat.transpose()
    tw[:, :, 1] = tdot.transpose() * weight

    # make matrix T*WT
    twt = np.dot(tw[:, :, 0], tmat) + np.dot(tw[:, :, 1], tdot)
    tw = tw.reshape(n_poly, 2 * n_points)

    # insert matrix T*W in matrix C2
    c2 = np.zeros([n_poly + 4, 2 * n_points])
    c2[0:n_poly] = tw
    c2[n_poly, 0] = 1
    c2[n_poly + 1, 1] = 1
    c2[n_poly + 2, -2] = 1
    c2[n_poly + 3, -1] = 1

    # insert matrix T*WT in matrix C1
    c1 = np.zeros([n_poly + 4, n_poly + 4])
    c1[0:n_poly, 0:n_poly] = twt
    c1[n_poly + 0, 0:n_poly] = tmat[0]
    c1[n_poly + 1, 0:n_poly] = tdot[0]
    c1[n_poly + 2, 0:n_poly] = tmat[-1]
    c1[n_poly + 3, 0:n_poly] = tdot[-1]

    c1[0:n_poly, n_poly:] = c1[n_poly:, 0:n_poly].transpose()

    c1inv = np.linalg.inv(c1)
    c1c2 = np.dot(c1inv, c2)
    c1c2 = c1c2.reshape(n_poly + 4, n_points, 2)
    c1c2 = c1c2[:, ::-1, :]
    c1c2 = c1c2.reshape(n_poly + 4, 2 * n_points)

    # separate even rows for x, and odd rows for dx/dt
    return c1c2[0:n_poly, 0::2], c1c2[0:n_poly, 1::2]


def make_cheb_matrix_only_x(n_points, n_poly):
    """Compute C1^(-1)C2 using Newhall89 approach without dx/dt

    Compute xMultiplier using only the equality constraint of the x-values
    at the endpoints.
    To be used when first derivatives are not available.
    If chebyshev approximations are strung together piecewise only the x-values
    and not the first derivatives will be continuous at the boundaries.
    Multiplying this matrix by the x-values to be fit produces the chebyshev
    coefficients. This function need only be called once for a given
    polynomial degree and
    number of points.
    See Newhall, X. X. 1989, Celestial Mechanics, 45, p. 305-310.

    Parameters
    ----------
    n_points : `int`
        Number of point to be fits. Must be greater than 2.
    n_poly : `int`
        Number of polynomial terms. Polynomial degree + 1

    Returns
    -------
    c1c2: `np.ndarray`
        xMultiplier, Even rows of C1^(-1)C2 w/ shape (n_points+1)x(n_poly)
        to be multiplied by x values
    """

    tmat = np.zeros([n_points, n_poly])
    cj = np.zeros([n_poly])
    xj = np.linspace(1, -1, n_points)
    for i in range(0, n_poly):
        cj[:] = 0
        cj[i] = 1
        tmat[:, i], v = chebeval(xj, cj)

    # Augment matrix T to get matrix C2
    c2 = np.zeros([n_poly + 2, n_points])
    c2[0:n_poly] = tmat.transpose()
    c2[n_poly, 0] = 1
    c2[n_poly + 1, n_points - 1] = 1

    # Augment matrix T*WT to get the matrix C1
    c1 = np.zeros([n_poly + 2, n_poly + 2])
    c1[0:n_poly, 0:n_poly] = np.dot(tmat.transpose(), tmat)
    c1[n_poly + 0, 0:n_poly] = tmat[0]
    c1[n_poly + 1, 0:n_poly] = tmat[-1]
    c1[0:n_poly, n_poly:] = c1[n_poly:, 0:n_poly].transpose()

    c1inv = np.linalg.inv(c1)
    # C1^(-1) C2
    c1c2 = np.dot(c1inv, c2)

    c1c2 = c1c2.reshape(n_poly + 2, n_points)
    c1c2 = c1c2[:, ::-1]
    return c1c2[0:n_poly]


def chebfit(t, x, dxdt=None, x_multiplier=None, dx_multiplier=None, n_poly=7):
    """Fit Chebyshev polynomial constrained at endpoints using
    Newhall89 approach.

    Return Chebyshev coefficients and statistics from fit
    to array of positions (x) and optional velocities (dx/dt).
    If both the function and its derivative are specified, then the value and
    derivative of the interpolating polynomial at the
    endpoints will be exactly equal to the input endpoint values.
    Many approximations may be piecewise strung together and the function value
    and its first derivative will be continuous across boundaries.
    If derivatives are not provided, only the function value will be
    continuous across boundaries.

    If x_multiplier and dx_multiplier are not provided or
    are an inappropriate shape for t and x, they will be recomputed.
    See Newhall, X. X. 1989, Celestial Mechanics, 45, p. 305-310
    for details.

    Parameters
    ----------
    t : `np.ndarray`
        Array of regularly sampled independent variable (e.g. time)
    x : `np.ndarray`
        Array of regularly sampled dependent variable (e.g. declination)
    dxdt : `np.ndarray', optional
        Optionally, array of first derivatives of x with respect to t,
        at the same grid points. (e.g. sky velocity ddecl/dt)
    x_multiplier : `np.ndarray`, optional
        Optional 2D Matrix with rows of C1^(-1)C2 corresponding to x.
        Use make_cheb_matrix to compute
    dx_multiplier : `np.ndarray`, optional
        Optional 2D Matrix with rows of C1^(-1)C2 corresponding to dx/dt.
        Use make_cheb_matrix to compute
    n_poly : `int`, optional
        Number of polynomial terms. Degree + 1.  Must be >=2 and <=2*n_points,
        when derivative information is specified, or <=n_points, when no
        derivative information is specified. Default = 7.

    Returns
    -------
    a_n : `np.ndarray`
        Array of chebyshev coefficients with length=n_poly.
    residuals : `np.ndarray`
        Array of residuals of the tabulated function x minus the
        approximated function.
    rms : `float`
        The rms of the residuals in the fit.
    maxresid : `float`
        The maximum of the residals to the fit.
    """
    n_points = len(t)
    if len(x) != n_points:
        raise ValueError("length of x (%s) != length of t (%s)" % (len(x), n_points))
    if dxdt is None:
        if n_poly > n_points:
            raise RuntimeError(
                "Without velocity constraints, n_poly (%d) must be less than %s" % (n_poly, n_points)
            )
        if n_poly < 2:
            raise RuntimeError("Without velocity constraints, n_poly (%d) must be greater than 2" % n_poly)
    else:
        if n_poly > 2 * n_points:
            raise RuntimeError(
                "n_poly (%d) must be less than %s (%d)" % (n_poly, "2 * n_points", 2 * (n_points))
            )
        if n_poly < 4:
            raise RuntimeError("n_poly (%d) must be greater than 4" % n_poly)

    # Recompute C1invX2 if x_multiplier and dx_multiplier are None or
    # they are not appropriate for sizes of input positions and velocities.

    if x_multiplier is None:
        redo_x = True
    else:
        redo_x = (x_multiplier.shape[1] != n_points) | (x_multiplier.shape[0] != n_poly)

    if dx_multiplier is None:
        redo_v = True
    else:
        redo_v = (dx_multiplier.shape[1] != n_points) | (dx_multiplier.shape[0] != n_poly)

    if (dxdt is None) & redo_x:
        x_multiplier = make_cheb_matrix_only_x(n_points, n_poly)

    if (dxdt is not None) & (redo_v | redo_x):
        x_multiplier, dx_multiplier = make_cheb_matrix(n_points, n_poly)

    if x.size != n_points:
        raise RuntimeError("Not enough elements in X")

    t_interval = np.array([t[0], t[-1]]) - t[0]
    t_scaled = t - t[0]

    # Compute the X portion of the coefficients
    a_n = np.dot(x_multiplier, x)

    # Compute statistics
    # for x and dxdt if it is available
    if dxdt is not None:
        a_n = a_n + np.dot(dx_multiplier, dxdt * (t_interval[1] - t_interval[0]) / 2.0)
        x_approx, dx_approx = chebeval(t_scaled, a_n, interval=t_interval)
    else:
        # Statistics for x only
        x_approx, _ = chebeval(t_scaled, a_n, interval=t_interval, do_velocity=False)

    residuals = x - x_approx
    se = np.sum(residuals**2)
    rms = np.sqrt(se / (n_points - 1))
    maxresid = np.max(np.abs(residuals))

    return a_n, residuals, rms, maxresid



__all__ = ("all_sky_db", "diode_sky_db")

import os

import numpy as np
import sqlalchemy as sqla
from rubin_scheduler.data import get_data_dir

# Tools for using an all-sky sqlite DB with cannon
# and photodiode data from the site.


def all_sky_db(date_id, sql_q=None, dtypes=None, db_address=None, filt="R"):
    """
    Fetch star and sky magnitudes from a processed all-sky sqlite database.

    Parameters
    ----------
    date_id : `float`
        Date (MJD) to fetch star observation information from the database.
    sql_q : `str`
        Sql query to use. None will use a default query to get all star info.
    dtypes : `list` [`str`, `dtype`]
        Data types expected from the database. None will use the defaults.
    db_address : `str`
        Database data path. Default uses db in $RUBIN_SIM_DATA/skybrightness.
    filt : `str`
        Filter in which to fetch stellar observation data.

    Returns
    -------
    data : `np.ndarray`, (N,)
        Stellar observation data.
    mjd : `float`
        MJD of the observations.
    """
    if db_address is None:
        data_path = os.path.join(get_data_dir(), "skybrightness")
        db_address = "sqlite:///" + os.path.join(data_path, "photometry", "skydata.sqlite")
    if sql_q is None:
        sql_q = (
            "select stars.ra, stars.dec,  obs.alt, obs.starMag, obs.sky, obs.filter from obs, "
            'stars where obs.starID = stars.ID and obs.filter = "%s" and obs.dateID = %i;' % (filt, date_id)
        )
    if dtypes is None:
        names = ["ra", "dec", "alt", "starMag", "sky", "filter"]
        types = [float, float, float, float, float, "|S1"]
        dtypes = list(zip(names, types))

    engine = sqla.create_engine(db_address)
    connection = engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(sql_q)
    data = cursor.fetchall()
    data = np.asarray(data, dtype=dtypes)

    q2 = "select mjd from dates where ID = %i" % date_id
    cursor.execute(q2)

    mjd = cursor.fetchall()
    if len(mjd) == 0:
        mjd = None
    else:
        mjd = mjd[0][0]
    return data, mjd


def diode_sky_db(mid_mjd, sql_q=None, dtypes=None, db_address=None, clean=True):
    """Fetch diode measurements of skybrightness."""
    if db_address is None:
        data_path = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        db_address = "sqlite:///" + os.path.join(data_path, "photometry", "skydata.sqlite")
    if sql_q is None:
        sql_q = "select mjd, R, Y, Z from photdiode where mjd > %f-1 and  mjd < %f+1" % (
            mid_mjd,
            mid_mjd,
        )
    if dtypes is None:
        names = ["mjd", "r", "y", "z"]
        types = [float] * 4
        dtypes = list(zip(names, types))

    engine = sqla.create_engine(db_address)
    connection = engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(sql_q)
    data = cursor.fetchall()
    data = np.asarray(data, dtype=dtypes)

    if clean:
        data = data[np.where((data["r"] > 0) & (data["z"] > 0) & (data["y"] > 0))]

    return data



import os
import shutil
import sqlite3
import tempfile
import unittest

import pandas as pd

import rubin_sim.maf.db as db


class TestTrackingDb(unittest.TestCase):
    def setUp(self):
        self.run_name = "test)_run"
        self.run_group = "test"
        self.opsim_comment = "runcomment"
        self.maf_comment = "mafcomment"
        self.maf_dir = "mafdir"
        self.maf_version = "1.0"
        self.maf_date = "2017-01-01"
        self.run_version = "4.0"
        self.run_date = "2017-02-01"
        self.db_file = None

    def test_test_tracking_db_creation(self):
        """Test tracking database creation."""
        tempdir = tempfile.mkdtemp(prefix="trackDb")
        tracking_db_file = os.path.join(tempdir, "tracking.db")
        trackingdb = db.TrackingDb(database=tracking_db_file)
        self.assertTrue(os.path.isfile(tracking_db_file))
        trackingdb.close()
        shutil.rmtree(tempdir)

    def test_test_add_run(self):
        """Test adding a run to the tracking database."""
        tempdir = tempfile.mkdtemp(prefix="trackDb")
        tracking_db_file = os.path.join(tempdir, "tracking.db")
        trackingdb = db.TrackingDb(database=tracking_db_file)
        track_id = trackingdb.add_run(
            run_group=self.run_group,
            run_name=self.run_name,
            run_comment=self.opsim_comment,
            run_version=self.run_version,
            run_date=self.run_date,
            maf_comment=self.maf_comment,
            maf_dir=self.maf_dir,
            maf_version=self.maf_version,
            maf_date=self.maf_date,
            db_file=self.db_file,
        )
        con = sqlite3.connect(tracking_db_file)
        res = pd.read_sql("select * from runs", con).to_records()
        self.assertEqual(res["maf_run_id"][0], track_id)
        # Try adding this run again. Should return previous track_id.
        track_id2 = trackingdb.add_run(maf_dir=self.maf_dir)
        self.assertEqual(track_id, track_id2)
        # Test will add additional run, with new track_id.
        track_id3 = trackingdb.add_run(maf_dir="test2")
        self.assertNotEqual(track_id, track_id3)
        trackingdb.close()
        con.close()
        shutil.rmtree(tempdir)

    def test_test_del_run(self):
        """Test removing a run from the tracking database."""
        tempdir = tempfile.mkdtemp(prefix="trackDb")
        tracking_db_file = os.path.join(tempdir, "tracking.db")
        trackingdb = db.TrackingDb(database=tracking_db_file)
        track_id = trackingdb.add_run(maf_dir=self.maf_dir)
        track_id2 = trackingdb.add_run(maf_dir=self.maf_dir + "test2")
        con = sqlite3.connect(tracking_db_file)
        res = pd.read_sql("select * from runs", con).to_records(index=False)
        self.assertEqual(res["maf_run_id"][0], track_id)
        # Test removal works.
        trackingdb.delRun(track_id)
        res = pd.read_sql("select * from runs", con).to_records(index=False)
        # The run returned here is track_id2
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0][0], track_id2)
        # Test cannot remove run which does not exist.
        self.assertRaises(Exception, trackingdb.delRun, track_id)
        trackingdb.close()
        con.close()
        shutil.rmtree(tempdir)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from rubin_scheduler.data import data_versions, get_data_dir

from rubin_sim.data import get_baseline
from rubin_sim.data import get_data_dir as gdd


class DataTest(unittest.TestCase):
    def testBaseline(self):
        """
        Get the baseline sim location
        """
        data_dir = get_data_dir()
        dd2 = gdd()

        assert data_dir == dd2

        if "sim_baseline" in os.listdir(data_dir):
            _ = get_baseline()
        _ = data_versions()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Takes a given set of galactic coordinates and a stellar
# mass range, then calculates the number of stars within that range
# that will be fainter than mag 16, and have sufficiently low noise
# in the given band. For use with Field Star Count metric


import numpy as np
from scipy.optimize import newton

from . import abs_mag, spec_type
from .starcount import starcount

xi = 1.0
alpha = 2.35


def IMF(lower, upper):
    exp = alpha - 1.0
    part1 = xi / exp
    part2 = lower**-exp - upper**-exp
    part3 = part1 * part2
    return part3


def get_distance(apparent, absolute):
    part1 = (apparent - absolute) / 5.0
    part2 = 10 * 10**part1
    return part2


def noise_opt(m, band, sigma):
    part1 = mag_error(m, band)
    part2 = part1**2.0
    part3 = sigma**2.0
    total = part2 - part3
    return total


def mag_error(ap_mag, band, calcm5=0):  # apparent magnitude and band
    gamma = {"u": 0.037, "g": 0.038, "r": 0.039, "i": 0.039, "z": 0.040, "y": 0.040}
    m5 = {"u": 23.9, "g": 25.0, "r": 24.7, "i": 24.0, "z": 23.3, "y": 22.1}
    if calcm5 == 0:
        calcm5 = m5[band]
    X = 10.0 ** (0.4 * (ap_mag - calcm5))
    random_2 = np.sqrt((0.04 - gamma[band]) * X + gamma[band] * X * X)
    error_2 = random_2**2.0 + (0.005) ** 2.0
    mag_error = np.sqrt(error_2)
    return mag_error


def noise_calc(band):
    gamma = {"u": 0.037, "g": 0.038, "r": 0.039, "i": 0.039, "z": 0.040, "y": 0.040}
    m5 = {"u": 23.9, "g": 25.0, "r": 24.7, "i": 24.0, "z": 23.3, "y": 22.1}
    sigma = 0.03
    sigma_sys = 0.005

    def fun(x):
        sigma_sys**2
        +(0.04 - gamma[band]) * 10 ** (0.4 * (x - m5[band]))
        +gamma[band] * 10 ** (0.8 * (x - m5[band]))
        -(sigma**2)

    return newton(fun, 25)


def dist_calc(mass, band):
    # mass to spectral type
    # spectral type to absolute mag
    bands = ["z", "y", "i", "r", "g", "u"]
    output = abs_mag.abs_mag(spec_type.spec_type(mass))[0]
    indexvalue = bands.index(band)
    absolutemag = output[indexvalue][0]
    apparent = noise_calc(band)
    dist_min = get_distance(16, absolutemag)
    dist_max = get_distance(apparent, absolutemag)
    return dist_min, dist_max
    # abs mag to apparent mag ranges, > 16, noise dependent upper limit


def starcount_bymass(eq_ra, eq_dec, m1, m2, band):
    masses = np.linspace(m1, m2, num=20)
    totmass = IMF(m1, m2)
    totmass = IMF(0.2, 1.04)
    massbins = IMF(masses[:-1], masses[1:])
    massfractions = massbins / totmass
    distances = [dist_calc(x, band) for x in masses[:-1]]
    starcounts = [y * starcount(eq_ra, eq_dec, x[0], x[1]) for x, y in zip(distances, massfractions)]
    return sum(starcounts)



"""Sets of metrics to look at general sky coverage -
nvisits/coadded depth/Teff.
"""

__all__ = ("nvisitsM5Maps", "tEffMetrics", "nvisitsPerNight", "nvisitsPerSubset")

import copy

import numpy as np

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as stackers
import rubin_sim.maf.utils as mafUtils

from .col_map_dict import col_map_dict
from .common import filter_list, standard_summary


def nvisitsM5Maps(
    colmap=None,
    runName="opsim",
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
    runLength=10.0,
):
    """Generate maps of the number of visits and coadded depth
    (with and without dust extinction) in all bands and per filter.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").
    slicer :  `rubin_sim.maf.slicer` or None, optional
        Optionally, use something other than an nside=64 healpix slicer
    runLength : `float`, optional
        Length of the simulated survey, for scaling values for the plot limits.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    subgroup = extraInfoLabel
    if subgroup is None:
        subgroup = "All visits"

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]
    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )
    # Set up some values to make nicer looking plots.
    benchmarkVals = mafUtils.scale_benchmarks(runLength, benchmark="design")
    # Check that nvisits is not set to zero (for very short run length).
    for f in benchmarkVals["nvisits"]:
        if benchmarkVals["nvisits"][f] == 0:
            print("Updating benchmark nvisits value in %s to be nonzero" % (f))
            benchmarkVals["nvisits"][f] = 1
    benchmarkVals["coaddedDepth"] = mafUtils.calc_coadded_depth(
        benchmarkVals["nvisits"], benchmarkVals["singleVisitDepth"]
    )
    # Scale the n_visit ranges for the runLength.
    nvisitsRange = {
        "u": [20, 80],
        "g": [50, 150],
        "r": [100, 250],
        "i": [100, 250],
        "z": [100, 300],
        "y": [100, 300],
        "all": [700, 1200],
    }
    scale = runLength / 10.0
    for f in nvisitsRange:
        for i in [0, 1]:
            nvisitsRange[f][i] = int(np.floor(nvisitsRange[f][i] * scale))

    # Generate Nvisit maps in all and per filters
    displayDict = {"group": "Nvisits Maps", "subgroup": subgroup}
    metric = metrics.CountMetric(colmap["mjd"], metric_name="NVisits", units="")

    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=64, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)
        slicerDust = slicers.HealpixSlicer(
            nside=64,
            lat_col=decCol,
            lon_col=raCol,
            lat_lon_deg=degrees,
            use_cache=False,
        )
    else:
        # If there is already a slicer set up, ensure we have one for dust
        # which is NOT using cache.
        slicerDust = copy.deepcopy(slicer)
        slicerDust.use_cache = False

    for f in filterlist:
        sql = sqls[f]
        displayDict["caption"] = f"Number of visits per healpix in {info_label[f]}."
        displayDict["order"] = orders[f]
        bin_size = 2
        if f == "all":
            bin_size = 5
        plotDict = {
            "x_min": nvisitsRange[f][0],
            "x_max": nvisitsRange[f][1],
            "color_min": nvisitsRange[f][0],
            "color_max": nvisitsRange[f][1],
            "bin_size": bin_size,
            "color": colors[f],
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standard_summary(),
        )
        bundleList.append(bundle)

    # Generate Coadded depth maps per filter
    displayDict = {"group": "Coadded M5 Maps", "subgroup": subgroup}
    metric = metrics.Coaddm5Metric(m5_col=colmap["fiveSigmaDepth"], metric_name="CoaddM5")

    for f in filterlist:
        # Skip "all" for coadded depth.
        if f == "all":
            continue
        sql = sqls[f]
        displayDict["caption"] = f"Coadded depth per healpix in {info_label[f]}."
        displayDict["caption"] += " More positive numbers indicate fainter limiting magnitudes."
        displayDict["order"] = orders[f]
        plotDict = {
            "percentile_clip": 98,
            "color": colors[f],
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standard_summary(),
        )
        bundleList.append(bundle)

    # Add Coadded depth maps per filter WITH extragalactic extinction added
    displayDict = {"group": "Extragalactic Coadded M5 Maps", "subgroup": subgroup}
    metric = metrics.ExgalM5(m5_col=colmap["fiveSigmaDepth"], metric_name="Exgal_CoaddM5")

    for f in filterlist:
        # Skip "all" for coadded depth.
        if f == "all":
            continue
        sql = sqls[f]
        displayDict["caption"] = (
            "Coadded depth per healpix for extragalactic purposes "
            "(i.e. combined with dust extinction maps), "
            f"in {info_label[f]}."
        )
        displayDict["caption"] += " More positive numbers indicate fainter limiting magnitudes."
        displayDict["order"] = orders[f]
        plotDict = {
            "percentile_clip": 90,
            "color": colors[f],
        }
        bundle = mb.MetricBundle(
            metric,
            slicerDust,
            sql,
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standard_summary(),
        )
        bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def tEffMetrics(
    colmap=None,
    runName="opsim",
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
):
    """Generate a series of Teff metrics.
    Teff total, per night, and sky maps (all and per filter).

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").
    slicer : `rubin_sim.maf.BaseSlicer` or None, optional
        Optionally, use something other than an nside=64 healpix slicer

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    subgroup = extraInfoLabel
    if subgroup is None:
        subgroup = "All visits"

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]

    if slicer is not None:
        skyslicer = slicer
    else:
        skyslicer = slicers.HealpixSlicer(nside=64, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )
    if info_label["all"] is None:
        info_label["all"] = "All visits"

    # Total Teff and normalized Teff.
    displayDict = {"group": "T_eff Summary", "subgroup": subgroup}
    displayDict["caption"] = "Total effective time of the survey (see Teff metric)."
    displayDict["order"] = 0
    metric = metrics.SumMetric(col="t_eff", metric_name="Total Teff")
    slicer = slicers.UniSlicer()
    bundle = mb.MetricBundle(
        metric,
        slicer,
        constraint=sqls["all"],
        display_dict=displayDict,
        info_label=info_label["all"],
    )
    bundleList.append(bundle)

    displayDict["caption"] = "Normalized total effective time of the survey (see Teff metric)."
    displayDict["order"] = 1
    metric = metrics.MeanMetric(col="t_eff", metric_name="Normalized Teff")
    normalized_teff_stacker = stackers.TeffStacker(normed=True)
    slicer = slicers.UniSlicer()
    bundle = mb.MetricBundle(
        metric,
        slicer,
        constraint=sqls["all"],
        stacker_list=[normalized_teff_stacker],
        display_dict=displayDict,
        info_label=info_label["all"],
    )
    bundleList.append(bundle)

    # Generate Teff maps in all and per filters
    displayDict = {"group": "T_eff Maps", "subgroup": subgroup}

    metric = metrics.MeanMetric(col="t_eff", metric_name="Normalized Teff")
    normalized_teff_stacker = stackers.TeffStacker(normed=True)
    for f in filterlist:
        displayDict["caption"] = "Normalized effective time of the survey, for %s" % info_label[f]
        displayDict["order"] = orders[f]
        plotDict = {"color": colors[f]}
        bundle = mb.MetricBundle(
            metric,
            skyslicer,
            sqls[f],
            stacker_list=[normalized_teff_stacker],
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standard_summary(),
        )
        bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def nvisitsPerNight(
    colmap=None,
    runName="opsim",
    binNights=1,
    extraSql=None,
    extraInfoLabel=None,
    subgroup=None,
):
    """Count the number of visits per night through the survey.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey. Default is "opsim".
    binNights : `int`, optional
        Number of nights to count in each bin.
    extraSql : `str` or None, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str` or None, optional
        Additional info_label to add before any below (i.e. "WFD").
    subgroup : `str` or None, optional
        Use this for the 'subgroup' in the display_dict, instead of info_label.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()

    subgroup = subgroup
    if subgroup is None:
        subgroup = extraInfoLabel
        if subgroup is None:
            subgroup = "All visits"

    infoCaption = extraInfoLabel
    if extraInfoLabel is None:
        if extraSql is not None:
            infoCaption = extraSql
        else:
            infoCaption = "all visits"

    bundleList = []

    displayDict = {"group": "Nvisits Per Night", "subgroup": subgroup}
    displayDict["caption"] = "Number of visits per night for %s." % (infoCaption)
    displayDict["order"] = 0
    metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits")
    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=binNights)
    bundle = mb.MetricBundle(
        metric,
        slicer,
        extraSql,
        info_label=infoCaption,
        display_dict=displayDict,
        summary_metrics=standard_summary(),
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def nvisitsPerSubset(
    colmap=None,
    runName="opsim",
    binNights=1,
    constraint=None,
    footprintConstraint=None,
    extraInfoLabel=None,
):
    """Look at the distribution of a given sql constraint or
    footprint constraint's visits, total number and distribution over time
    (# per night), if possible.

    Parameters
    ----------
    opsdb : `str` or database connection
        Name of the opsim sqlite database.
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    binNights : `int`, optional
        Number of nights to count in each bin.
    constraint : `str` or None, optional
        SQL constraint to add to all metrics.
        This would be the way to select only a given "Note".
    footprintConstraint : `np.ndarray` or None, optional
        Footprint to look for visits within
        (and then identify via WFDlabelStacker).
        The footprint = a full length heapix array, filled with 0/1 values.
    extraInfoLabel : `str` or None, optional
        Additional info_label to add before any below (i.e. "WFD").

    Returns
    -------
    metric_bundleDict : `dict` of `rubin_sim.maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()

    bdict = {}
    bundleList = []

    if footprintConstraint is None:
        if extraInfoLabel is None and constraint is not None:
            extraInfoLabel += " %s" % constraint
        # Nvisits per night, this constraint.
        bdict.update(
            nvisitsPerNight(
                colmap=colmap,
                runName=runName,
                binNights=binNights,
                extraSql=constraint,
                extraInfoLabel=extraInfoLabel,
            )
        )
        # Nvisits total, this constraint.
        metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits")
        slicer = slicers.UniSlicer()
        displayDict = {
            "group": "Nvisit Summary",
            "subgroup": extraInfoLabel,
        }
        displayDict["caption"] = f"Total number of visits for {extraInfoLabel}."
        bundle = mb.MetricBundle(
            metric,
            slicer,
            constraint,
            info_label=extraInfoLabel,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    # Or count the total number of visits that contribute
    # towards a given footprint
    if footprintConstraint is not None:
        # Set up a stacker to use this footprint to label visits
        if extraInfoLabel is None:
            extraInfoLabel = "Footprint"
        footprintStacker = stackers.WFDlabelStacker(
            footprint=footprintConstraint,
            fp_threshold=0.4,
            area_id_name=extraInfoLabel,
            exclude_dd=True,
        )
        metric = metrics.CountSubsetMetric(
            col="area_id", subset=extraInfoLabel, units="#", metric_name="Nvisits"
        )
        slicer = slicers.UniSlicer()
        displayDict = {
            "group": "Nvisit Summary",
            "subgroup": extraInfoLabel,
            "caption": f"Visits within footprint {extraInfoLabel}.",
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            constraint,
            stacker_list=[footprintStacker],
            info_label=extraInfoLabel,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    for b in bundleList:
        b.set_run_name(runName)
    bdict.update(mb.make_bundles_dict_from_list(bundleList))
    return bdict



###################################################################
# Metric to evaluate the transientTimeSamplingMetric
#
# Author - Rachel Street: rstreet@lco.global
###################################################################
__all__ = (
    "calc_interval_decay",
    "GalPlaneVisitIntervalsTimescaleMetric",
    "GalPlaneSeasonGapsTimescaleMetric",
)

from types import MethodType

import numpy as np
from rubin_scheduler.utils import calc_season

from rubin_sim.maf.maps.galactic_plane_priority_maps import gp_priority_map_components_to_keys

from .base_metric import BaseMetric
from .galactic_plane_metrics import galplane_priority_map_thresholds
from .season_metrics import find_season_edges

TAU_OBS = np.array([2.0, 5.0, 11.0, 20.0, 46.5, 73.0])


def calc_interval_decay(delta_tobs, tau):
    # Decay constant for metric value relationship as function of obs interval
    K = 1.0 / tau
    m = np.exp(-K * (delta_tobs - tau))
    # But where observation interval is <= tau, replace with 1
    m[np.where(delta_tobs <= tau)] = 1.0
    return m


# this is a bit of a hack .. it helps us use a variety of tau_obs values,
# and dynamically set up reduce functions
def help_set_reduce_func(obj, metricval, tau):
    def _reduce_tau(obj, metricval):
        return metricval[tau]

    return _reduce_tau


class GalPlaneVisitIntervalsTimescaleMetric(BaseMetric):
    """Evaluate the intervals between sequential observations in a
    lightcurve relative to the scientifically desired sampling interval.

    Parameters
    ----------
    science_map : `str`
        Name of the priority footprint map key to use from the column
        headers contained in the priority_GalPlane_footprint_map_data tables.
    tau_obs : `np.ndarray` or `list` of `float`, opt
        Timescales of minimum-required observations intervals for various
        classes of time variability.
        Default (None), uses TAU_OBS. In general, this should be left as the
        default and consistent across all galactic-plane oriented metrics.
    mag_limit : `float`, opt
        Magnitude limit to use as a cutoff for various observations.
        Default 22.0.
    mjd_col : `str`, opt
        The name of the observation start MJD column.
        Default 'observationStartMJD'.
    m5_col : `str', opt
        The name of the five sigma depth column. Default 'fiveSigmaDepth'.
    """

    def __init__(
        self,
        science_map,
        tau_obs=None,
        mag_limit=22.0,
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        **kwargs,
    ):
        self.science_map = science_map
        self.priority_map_threshold = galplane_priority_map_thresholds(self.science_map)
        # tau_obs is an array of minimum-required observation intervals for
        # four categories of time variability
        if tau_obs is not None:
            self.tau_obs = tau_obs
        else:
            self.tau_obs = TAU_OBS
        # Create reduce functions for the class that are return the metric
        # for each value in tau_obs

        self.mag_limit = mag_limit
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        maps = ["GalacticPlanePriorityMap"]
        if "metric_name" not in kwargs:
            metric_name = f"GalPlaneVisitIntervalsTimescales_{self.science_map}"
        else:
            metric_name = kwargs["metric_name"]
            del kwargs["metric_name"]
        for tau in self.tau_obs:
            tau_reduce_name = f"reduce_Tau_{tau:.1f}".replace(".", "_")
            newmethod = help_set_reduce_func(self, None, tau)
            setattr(self, tau_reduce_name, MethodType(newmethod, tau_reduce_name))
        super().__init__(
            col=[self.mjd_col, self.m5_col],
            metric_name=metric_name,
            maps=maps,
            **kwargs,
        )
        for i, tau in enumerate(self.tau_obs):
            self.reduce_order[f"reduceTau_{tau:.1f}".replace(".", "_").replace("reduce", "")] = i

    def run(self, data_slice, slice_point=None):
        # Check if we want to evaluate this part of the sky,
        # or if the weight is below threshold.
        if (
            slice_point[gp_priority_map_components_to_keys("sum", self.science_map)]
            <= self.priority_map_threshold
        ):
            return self.badval
        # Select observations in the time sequence that fulfill the
        # S/N requirements:
        match = np.where(data_slice[self.m5_col] >= self.mag_limit)[0]
        # We need at least two visits which match these requirements
        # to calculate visit gaps
        if len(match) < 2:
            return self.badval
        # Find the time gaps between visits (in any filter)
        times = data_slice[self.mjd_col][match]
        times.sort()
        delta_tobs = np.diff(times)
        # Compare the time gap distribution to the time gap required
        # to characterize variability
        metric_data = {}
        for tau in self.tau_obs:
            # Normalize
            metric_data[tau] = calc_interval_decay(delta_tobs, tau).sum() / len(delta_tobs)
        return metric_data


class GalPlaneSeasonGapsTimescaleMetric(BaseMetric):
    """Evaluate the gap between sequential seasonal gaps in
    observations in a lightcurve relative to the scientifically desired
    sampling interval.

    Parameters
    ----------
    science_map : `str`
        Name of the priority footprint map key to use from the column
        headers contained in the
        priority_GalPlane_footprint_map_data tables.
    tau_var : `np.ndarray` or `list` of `float`, opt
        Timescales of variability for various classes of time variability.
        Default (None), uses TAU_OBS * 5. In general, this should be left
        as the default and consistent
        across all galactic-plane oriented metrics.
    mag_limit : `float`, opt
        Magnitude limit to use as a cutoff for various observations.
        Default 22.0.
    expected_season_gap : `float`, opt
        The typical season gap expected for a galactic plane field in days.
        The default, 145 days, is typical for a bulge field.
    mjd_col : `str`, opt
        The name of the observation start MJD column.
        Default 'observationStartMJD'.
    m5_col : `str', opt
        The name of the five sigma depth column. Default 'fiveSigmaDepth'.
    """

    def __init__(
        self,
        science_map,
        tau_var=None,
        mag_limit=22.0,
        expected_season_gap=145,
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        **kwargs,
    ):
        self.science_map = science_map
        self.priority_map_threshold = galplane_priority_map_thresholds(self.science_map)
        # tau_obs is an array of minimum-required observation intervals for
        # four categories of time variability; tau_var is the related timescale
        # for the variability (tau_var is approximately 5*tau_obs, in general)
        if tau_var is not None:
            self.tau_var = tau_var
        else:
            self.tau_var = TAU_OBS * 5
        ### NOTE: I would recommend dropping tau_var 10 and 25 from this
        # analysis unless the metric is changed
        # these intervals are so short they will *always* be dropped
        # during the season gap
        self.mag_limit = mag_limit
        self.expected_season_gap = expected_season_gap
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        if "metric_name" not in kwargs:
            metric_name = f"GalPlaneSeasonGapsTimescales_{self.science_map}"
        else:
            metric_name = kwargs["metric_name"]
            del kwargs["metric_name"]
        for tau in self.tau_var:
            tau_reduce_name = f"reduce_Tau_{tau:.1f}".replace(".", "_")
            newmethod = help_set_reduce_func(self, None, tau)
            setattr(self, tau_reduce_name, MethodType(newmethod, tau_reduce_name))
        super().__init__(col=[self.mjd_col, self.m5_col], metric_name=metric_name, **kwargs)
        for i, tau in enumerate(self.tau_var):
            self.reduce_order[f"reduce_Tau_{tau:.1f}".replace(".", "_").replace("reduce", "")] = i

    def run(self, data_slice, slice_point):
        # Check if we want to evaluate this part of the sky,
        # or if the weight is below threshold.
        if (
            slice_point[gp_priority_map_components_to_keys("sum", self.science_map)]
            <= self.priority_map_threshold
        ):
            return self.badval
        # Find the length of the gaps between each season
        times = data_slice[self.mjd_col]
        times.sort()
        # data = np.sort(data_slice[self.mjd_col], order=self.mjd_col)
        # SlicePoints ra/dec are always in radians -
        # convert to degrees to calculate season
        seasons = calc_season(np.degrees(slice_point["ra"]), times)
        first_of_season, last_of_season = find_season_edges(seasons)
        # season_lengths = times[last_of_season] -  times[first_of_season]
        # would this match interval calc better?
        season_gaps = times[first_of_season][1:] - times[last_of_season][:-1]
        if len(season_gaps) == 0:
            return self.badval
        metric_data = {}
        for i, tau in enumerate(self.tau_var):
            metric_data[tau] = calc_interval_decay(season_gaps, tau)
            # if the season gap is shorter than the expected season gap,
            # count this as 'good'
            good_season_gaps = np.where(season_gaps <= self.expected_season_gap)
            metric_data[tau][good_season_gaps] = 1
            metric_data[tau] = metric_data[tau].sum() / len(season_gaps)
        return metric_data



__all__ = ["NoOffset", "OffsetSNR", "BaseOffset"]

import numpy as np

from ..maf.utils.astrometry_utils import m52snr

# from lsst.sims.selfcal.clouds.Arma import ArmaSf, Clouds


class BaseOffset:
    """Base class for how to make offset classes"""

    def __init__(self, **kwargs):
        self.newkey = "dmag_keyname"
        pass

    def __call__(self, stars, visit, **kwargs):
        pass


class NoOffset(BaseOffset):
    def __init__(self):
        """Make no changes to the mags"""
        self.newkey = "dmag_zero"

    def __call__(self, stars, visits, **kwargs):
        dmag = np.zeros(stars.size, dtype=list(zip([self.newkey], [float])))
        return dmag


class OffsetSys(BaseOffset):
    def __init__(self, error_sys=0.003):
        """Systematic error floor for photometry"""
        self.error_sys = error_sys
        self.newkey = "dmag_sys"

    def __call__(self, stars, visits, **kwargs):
        nstars = np.size(stars)
        dmag = np.random.randn(nstars) * self.error_sys
        return dmag


class OffsetClouds(BaseOffset):
    """Offset based on cloud structure.
    Not used, as not fully implemented in this version (ArmaSf).
    """

    def __init__(self, sampling=256, fov=3.5):
        self.fov = fov
        self.newkey = "dmag_cloud"
        # self.SF = ArmaSf()
        self.SF = None
        # self.cloud = Clouds()
        self.cloud = None

    def __call__(self, stars, visits, **kwargs):
        # XXX-Double check extinction is close to the Opsim transparency
        extinc_mags = visits["transparency"]
        if extinc_mags != 0.0:
            sf_theta, sf_sf = self.SF.CloudSf(500.0, 300.0, 5.0, extinc_mags, 0.55)
            # Call the Clouds
            self.cloud.makeCloudImage(sf_theta, sf_sf, extinc_mags, fov=self.fov)
            # Interpolate clouds to correct position.
            # Nearest neighbor for speed?
            nim = self.cloud.cloudimage[0, :].size
            # calc position in cloud image of each star
            starx_interp = (np.degrees(stars["x"]) + self.fov / 2.0) * 3600.0 / self.cloud.pixscale
            stary_interp = (np.degrees(stars["y"]) + self.fov / 2.0) * 3600.0 / self.cloud.pixscale

            # Round off position and make it an int
            starx_interp = np.round(starx_interp).astype(int)
            stary_interp = np.round(stary_interp).astype(int)

            # Handle any stars that are out of the field for some reason
            starx_interp[np.where(starx_interp < 0)] = 0
            starx_interp[np.where(starx_interp > nim - 1)] = nim - 1
            stary_interp[np.where(stary_interp < 0)] = 0
            stary_interp[np.where(stary_interp > nim - 1)] = nim - 1

            dmag = self.cloud.cloudimage[starx_interp, stary_interp]
        else:
            dmag = np.zeros(stars.size)
        return dmag


class OffsetSNR(BaseOffset):
    """Generate offsets based on the 5-sigma limiting depth of an observation
    and the brightness of the star.

    Note that this takes into account previous offsets that have been applied
    (so run this after things like vignetting).
    """

    def __init__(self, lsst_filter="r"):
        self.lsst_filter = lsst_filter
        self.newkey = "dmag_snr"

    def calc_mag_errors(self, magnitudes, m5, err_only=False):
        """ """
        snr = m52snr(magnitudes, m5)
        # via good old https://www.eso.org/~ohainaut/ccd/sn.html
        magnitude_errors = 2.5 * np.log10(1.0 + 1.0 / snr)
        if err_only:
            dmag = magnitude_errors
        else:
            dmag = np.random.randn(len(magnitudes)) * magnitude_errors
        return dmag

    def __call__(self, stars, visit, dmags=None):
        if dmags is None:
            dmags = {}
        temp_mag = stars[self.lsst_filter + "mag"].copy()
        # calc what magnitude the star has when it hits the silicon.
        # Thus we compute the SNR noise
        # AFTER things like cloud extinction and vignetting.
        for key in list(dmags.keys()):
            temp_mag = temp_mag + dmags[key]
        dmag = self.calc_mag_errors(temp_mag, visit["fiveSigmaDepth"])
        return dmag



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestSchedviewMetrics(unittest.TestCase):
    def test_age_metric(self):
        data = np.rec.fromrecords([(1, 60000)], names="id,observationStartMJD")

        assert metrics.AgeMetric(60002).run(data) == 2
        assert metrics.AgeMetric(60002.5).run(data) == 2.5
        assert np.isnan(metrics.AgeMetric(70000.0).run(data))
